# -*- coding: utf-8 -*-

import sys
if not "tinypy" in sys.version:
    from extra import *

import base_building
import toolz

class ShipyardBuilding(base_building.BaseBuilding):
    def __init__(self, cdata):
        self.show_ship = False
        self.show_unit = False  # показывать ли юнита на верфи
        base_building.BaseBuilding.__init__(self, cdata)
        self.ser_name = 'ShipyardBuilding'
        self.SetIsHome(False)
        self.hack_for_upgrade = True
# ============================================================
    # @override
    def init_fsm(self):
        base_building.BaseBuilding.init_fsm(self)
        self.states[StateRepairing.name] = StateRepairing(self)
        self.states[StateUpgrading.name] = StateUpgrading(self)
# ============================================================
    # @override
    def switch_animation(self):
        base_building.BaseBuilding.switch_animation(self)
        self.Drawer.ShowImages('unit', self.show_unit)
        if self.show_ship:
            self.Drawer.ShowImages('ship', True)
# ============================================================
    # @override
    def on_message(self, msg_type, msg_param1, msg_param2, msg_dict):
        '''Обработчик всех обычных сообщений (которые не были обработаны в on_msg)'''
        if msg_type == 'hammer_event':
            self.PlaySound('SND_BOATYARD_HAMMER')
# ============================================================
    #@override
    def on_signal(self, signal):
        if signal == 'offscreen':
            if self.is_offscreen: return    # чтобы не дублировать
            self.is_offscreen = True
            if not self.has_next_level:  # достигнут максимальный вариант...
                self.switch_level_to_ship()
        elif signal == 'onscreen':
            self.is_offscreen = False
        else:
            base_building.BaseBuilding.on_signal(self, signal)
# ============================================================
#    #@override
#    def on_state_change(self, cur_state, next_state):
#        base_building.BaseBuilding.on_state_change(self, cur_state, next_state)
#        if cur_state != next_state:
#            if next_state == 'upgrade':       # затычка для особого поведения верфи, там анимация должна проигрываться
#                self.Drawer.RestartEffect('build')
#                self.Drawer.SetAnimationEvent('unit', 50, 'hammer_event')
# ============================================================
    #@override
    def on_building_create(self):
        base_building.BaseBuilding.on_building_create(self)
        self.is_deactivated = self.GetAttrBool('deactivated')
# ============================================================
    #@override
    def switch_level_to_ship(self):
        '''Здесь происходит "апгрейд" здания, смена варианта. Здесь должны (!) проверяться все необходимые требования, кроме кол-ва ресурсов'''
        var_name = "level_2"
        if var_name in self.variants_switch:
            next_var_name = self.variants_switch[var_name]
            trace('switch_level_to_ship:')
            trace(' switch_level_to_ship:', self.get('name'), self.get('uid'))
            self.reset_cache()
            self.SwitchLevel(next_var_name)   # переключаемся на следующий вариант
            self.update_next_res()
            self.show_ship = True
            self.switch_animation()
            self.unreserve_action_cells()
        self.update_res_enough_state()
# ============================================================
    #@override
    def scale_bump(self):
        if not self.show_unit:
            base_building.BaseBuilding.scale_bump(self)
# ============================================================
    #@override
    def _get_tips_extra(self, place, result, by_click):
        if place == 'before_name':
            if self.is_mouse_over_zone == 'ship':
                result.append( ("text", "COLOR_HEADER") )
                ship_text = self.GetAttrStr("ship_text")
                result.append( ("text", ship_text) )
                return False
        return True
# ============================================================
    #@override
    def Get_tip_position(self, tip_size, camera_pos, window_rect, zoom):
        '''Возвращает координаты типса от левого-верхнего угла картинки типса'''
        if self.is_mouse_over_zone == 'ship':
            return toolz.Tools.get_tip_position(self, tip_size, camera_pos, window_rect, 'notip_ship_rect')
        else:
            return toolz.Tools.get_tip_position(self, tip_size, camera_pos, window_rect)
# ============================================================
    #@override
    def _switch_animation_extra(self, place):
        if place == 'upgrade':
            self.Drawer.ShowImages('build')     # для верфи
            self.Drawer.SetProgressedAnimation('build')
# ============================================================
    #@override
    def get_action_text(self, for_task_slide = False):
        if not for_task_slide:
            return 'DLG_TASK_SHIPYARD_CONSTRUCT_FULL'
        else:
            return 'DLG_TASK_SHIPYARD_CONSTRUCT'
# ============================================================
# ============================================================
# ============================================================
class StateRepairing(base_building.StateRepairing):
    def __init__(self, item = None):
        base_building.StateRepairing.__init__(self, item)
        self.ser_name = 'ShipyardStateRepairing'

    #@override
    def on_enter(self, evID):
        base_building.StateRepairing.on_enter(self, evID)
        self.item.Drawer.RestartEffect('build')
        self.item.Drawer.SetAnimationEvent('unit', 50, 'hammer_event')
# ============================================================
class StateUpgrading(base_building.StateUpgrading):
    def __init__(self, item = None):
        base_building.StateUpgrading.__init__(self, item)
        self.ser_name = 'ShipyardStateUpgrading'

    #@override
    def on_enter(self, evID):
        base_building.StateUpgrading.on_enter(self, evID)
        self.item.Drawer.RestartEffect('build')
        self.item.Drawer.SetAnimationEvent('unit', 50, 'hammer_event')
# ============================================================
