# -*- coding: utf-8 -*-

import sys
if not "tinypy" in sys.version:
    from extra import *

import anim_helper
import task_res
import task_res_store
import task_upgrade
import task_callworker
import task_walkw
import task_classes
import fsm_state
import toolz

#@statemachine
class BaseBuilding(Building):
    #cancel_switch_place_variant_name = 'map_place'
    variants_switch = {
            'map_place': 'place',
            'map_place1': 'place',
            'map_place2': 'place',
            'map_place3': 'place',
            'place': 'level_1',
            'level_1': 'level_2',
            'level_2': 'level_3',
            'level_3': 'level_4',
            'level_4': 'level_5',
            'level_5': 'level_6',
            'level_6': 'level_7',
            'level_7': 'level_8',
            'level_8': 'level_9',
            'level_9': 'level_10',  # до 10 апгрейдов максимум
            'broken_1': 'level_1',
            'broken_2': 'level_2',
            'broken_3': 'level_3'
        }
    variants_indexes = {
            'map_place': 0,
            'map_place1': 0,
            'map_place2': 0,
            'map_place3': 0,
            'place': 0,
            'level_1': 1,
            'level_2': 2,
            'level_3': 3,
            'level_4': 4,
            'level_5': 5,
            'level_6': 6,
            'level_7': 7,
            'level_8': 8,
            'level_9': 9,  # до 10 апгрейдов максимум
            'broken_1': 0,
            'broken_2': 1,
            'broken_3': 2,
            'thrown_1': 0,
            'thrown_2': 1,
            'thrown_3': 2
        }
    blocked_switch = {
            'level_1': 'thrown_1',
            'level_2': 'thrown_2',
            'level_3': 'thrown_3'
        }
    broken_switch = {
            'level_1': 'broken_1',
            'level_2': 'broken_2',
            'level_3': 'broken_3'
        }
    resource_names = ['food', 'bricks', 'boards', 'coins', '_berries', '_fish', '_flour', '_wood', '_ore', '_stones', '_granite']
    res_images = {
        'food':'IMG_I_TIP_FOOD',
        'boards':'IMG_I_TIP_BOARDS',
        'bricks':'IMG_I_TIP_BRICKS',
        'coins':'IMG_I_TIP_COINS',
        'gold': 'IMG_I_TIP_GOLD',
        '_berries':'IMG_I_TIP_FOOD',
        '_wood':'IMG_I_TIP_WOOD',
        '_fish':'IMG_I_TIP_FISH',
        '_flour':'IMG_I_TIP_BAG',
        '_ore':'IMG_I_TIP_ORE',
        '_stones':'IMG_I_TIP_STONE',
        '_granite':'IMG_I_TIP_STONE',
        }
    shaded_layers = ['floor','default','unit','resources','upgrading','build','anim']
    bump_layers = ['default','unit','upgrading','build','anim']
    not_for_attack = {'hq':0, 'hq_small':0, 'watchtower':0, 'house':0,\
                        'boatyard':0, 'verf':0, 'bridge':0, 'column':0,\
                        'dam':0, 'enemy':0, 'farm_field':0, 'fence':0, 'no_fog':0,\
                        'ind_house_w':0, 'totem':0, 'totem_stone':0, 'forester_field':0,
                        'waterpump':0}
    has_no_effectiveness = {'hq':0, 'hq_small':0, 'watchtower':0, 'house':0,\
                        'boatyard':0, 'verf':0, 'bridge':0, 'column':0,\
                        'dam':0, 'enemy':0, 'fence':0, 'no_fog':0,\
                        'ind_house_w':0, 'totem':0, 'totem_stone':0,\
                        'warehouse':0, 'shoemaker':0, 'workshop':0,
                        'waterpump':0}      # 'woodcutter':0, 'quarry':0,
    fake_buildings = {'dam':0, 'bridge':0, 'enemy':0, 'fence':0, 'house_fake':0, 'no_fog':0,
                        'totem':0, 'totem_stone':0, }   # здания, которые используют тип но зданиями не являются, либо просто здания-пустышки
    no_button_states = {'wait_unit':0, 'repairing':0,
                        'upgrading':0, 'removing':0, 'removing_final':0, '_dead_':0,
                        'building':0, 'survived':0, 'map_place':0, 'init':0}

    builders_states = {'build_wait':0, 'building':0, 'upgrading':0, 'repairing':0,
                        'wait_unit':0, 'repair_wait':0}
    wait_states = ['build_wait', 'repair_wait']
    _non_cancelable_sates = ['building', 'upgrading', 'repairing']
    highlight_layers = ['floor','default','unit','anim','build','upgrading']

    button_active_frames = [0,3,6, 9,12,15, 18]
    cancel_build_active_frame = 3

    # trying to optimize a bit:
    _normal__empty = ['normal', '']
    _normal__broken = ['normal', 'broken']
    _place__broken = ['place', 'broken']
    _broken__place__map_place = {'broken':0, 'place':0, 'map_place':0}
    _map_place__place = ['map_place', 'place']
    _upgrade__build = ['upgrade', 'build']
# ============================================================
# ============================================================
    def __init__(self, cdata):
        Building.__init__(self, cdata)

        self.ser_name = '__none_basebuilding__'
        self.need_deserialize_init_extra = True
        self.uid = -1
        self.subtype = self.get('stype')
        self.prev_mplace_name = '__none__'
        self.states = {}
        self.is_upgrade_units_waiting = False
        self.start_cell_pos = {'col':0, 'row':0}        # ячейка, из/в которой будут выбегать юниты
        self.max_units_count = 0 # default - no units                  # максимальное число юнитов в здании
        self.cur_units_count = 0                    # текущее число юнитов данного здания
        self.external_units_count = 0               # сколько юнитов "считается" снаружи
        self.status = '__none__' #"place"                                  # статус  здания, обычно: [normal, broken, ...]
        self.prev_status = '__none__'               # предыдущий статус здания (используется при смене варианта здания)
        self.current_episode = self.GetEpisodeIndex()
        self.is_territory = self.GetAttrBool('is_territory')
        self.is_thrown = False
        self.need_max_time = False          # нужно ли выставить 'res_gen_max_time'
        self.max_res_generation_time = 0
        self.max_res_gen_time = 0
        self.has_effectivness_inited = False    # здание работало хотя бы сколько-то? Нужно, чтобы не сбрасывать эффективность полей после ремонта после атаки варваров
        self.current_effectiveness = 100
        self.restore_eff_k = 1.0
        self.generation_count = 0
        self.effectiveness_degradation_count = 0
        self.is_priority_building = False   # здание строится по приоритету
        self.build_time = 1                    # время, необходимое для постройки/апгрейда/ремонта здания
        self.build_progress = 0                     # сколько времени постройки уже прошло
        self.hack_for_upgrade = False               # здание апгрейдится внутренней анимацией (все yard'ы)
        self.anim_helper = anim_helper.AnimHelper()
        self.anim_helper.initialize(self.Drawer, self.Signals, 'ah_signal')
        self._building_units = []         # free units for actions  - юниты данного здания
        self.walking_home_units = []                # юниты, бегущие в здание (домой)
        self.max_build_units = self.GetAttrInt('max_build_units')
        self.worker_units = []          # юниты, которые работают в здании
        self.barb_cells_names = []     # названия атрибутов-ячеек для атаки здания дикарями
        self.building_cells_names = []     # названия атрибутов-ячеек для постройки здания
        self.reserved_building_cells_names = []   # зарезервированные для использования названия ячеек
        self.cell_pos = self.get('cell')        # ячейка, к которой привязано положение здания на карте - относительно нее считаются другие ячейки
        self.res_generation_time = 0           # время, затрачиваемое на генерацию ресурса
        self.is_generating_res = False          # генерируется ли ресурс
        self.is_generated_res = False           # ресурс сгенерирован и лежит возле здания?
        self.is_generated_res_doubled = False   # ресурс сгенерирован во время бонуса и он двойной
        self.to_building_subtype = '__none__'   # название подтипа здания, в которое нужно нести ресурсы от данного здания
        self.res_cell = {'col':0, 'row':0}      # ячейка выкладывания ресурса
        self.has_base_res = False       # достаточно ли базового ресурса, из которого идет производство
        self.tasks_member = {}          # участником каких заданий является данное здание. см. start_task_participation, stop_task_participation
        self.is_wait_take_res_unit = False  # здание ожидает юнита, который заберет ресурс
        self.is_wait_res_unit = False   # здание ожидает юнита с базовым ресурсом (для перерабатывающих зданий)
        self.has_next_level = False         # есть ли куда обновлять здание? Или это уже максимум?
        self.next_res = None                #??? кол-во ресурсов, необходимое для апгрейда до следующего варианта
        self.button_res_enough = False  # достаточно ли ресурсов для апгрейда - используется в состоянии кнопки
        self.is_zone_hlighted = False   # находится ли курсор над объектом, отличается от self.is_mouse_over, предназначен для показа зоны подсветки
        self.is_mouse_over = False          # мышь находится над объектом?
        self.is_selected = False            # объект выделен?
        self.is_indians = False             # это индейское здание?
        self.is_under_attack = 0    # на здание нападает варвар, со зданием ничего нельзя делать (счетчик, варваров может быть больше одного)
        self.energy = 100 #percent
        self.energy_hit = self.GetAttrInt('action_hit') #percent per hit
        if self.energy_hit == 0:
            self.energy_hit = 10
        self.is_internal_free_units = False     # если вместо 'free_unit' из UMgr нужно использовать "внутренний" список
        self._internal_free_units = []          # внутренний список свободных юнитов
        self.boost_work_percent = self.GetAttrInt('boost_work_percent')
        self.action_unit_name = self.GetAttrStr('action_unit_name')     # название шаблона юнита, который генерируется в данном здании
        self.max_variant = self.GetAttrStr('max_variant')               # ограничение на максимальный вариант при апгрейдах
        self.is_deactivated = self.GetAttrBool('deactivated')           # здание деактивировано для взаимодействия с игроком (типсы, похоже, лучше оставить)
        self.is_mouse_over_zone = ''                                    # название группы изображений, над которой сейчас находится курсор
        self.non_selectable = self.GetAttrBool('non_selectable')        # объект "невыделяемы", не подсвечиваемый и без селекта - почти как deactivated, но deactivated может активироваться, а вот non_selectable - никогда
        self.is_offscreen = True
        self.arrow_shown = False
        self.clock_shown = False
        self.lock_shown = False
        self.prev_fog_r = 0
        self.fog_r = self.GetAttrInt('fog_r')
        #self.task_call_res_uid = 0
        self.show_action_zone = False       # пока для Woodcutter - зона деревьев
        self.extra_event = self.GetAttrStr('extra_event')   # нужно в самом начале прочитать эти данные, чтобы вариант не перетер
        self.blocked_remove = False
        self.unblock_remove = False
        self.unblock_restore_eff = False
        self.has_pending_actions = False
        self.is_shaded = self.GetAttrBool('shaded')

        self.cached_next_level_name_check = '__reset__'
        self.cached_next_level_name_skipcheck = '__reset__'
        self.cached_menu_items = None
        self.cached_is_clickable_noskip = None
        self.cached_is_clickable_skip = None
        self.cached_need_remove_btn = None
        self.cached_need_update_skip_btn = None
        self.cached_need_update_noskip_btn = None
        self.cached_raw_green_state = None
        self.cached_menu_item_green = {}

        if self.is_shaded:
            self.Drawer.SetShaded(self.shaded_layers, True)
# ============================================================
    def extra_init(self):
        '''Вызывается при создании объекта. Используется при переопределении.'''
        if self.non_selectable:
            self.SetMouseActive(False)      # отключаем проверку на взаимодействие с мышью
        self.on_before_building_complete()
# ============================================================
    def dispose(self):
        pass
# ============================================================
    def deserialize_init_extra(self):
        self.anim_helper.drawer = self.Drawer
        self.anim_helper.signaler = self.Signals
# ============================================================
    def reset_cache(self):
        self.cached_next_level_name_check = '__reset__'
        self.cached_next_level_name_skipcheck = '__reset__'
        self.cached_menu_items = None
        self.cached_is_clickable_noskip = None
        self.cached_is_clickable_skip = None
        self.cached_need_remove_btn = None
        self.cached_need_update_skip_btn = None
        self.cached_need_update_noskip_btn = None
        self.cached_raw_green_state = None
        self.cached_menu_item_green = {}
# ============================================================
    def has_access(self):
        return self.HasAccess()
# ============================================================
    def set_under_attack(self):
        self.is_under_attack += 1
        if self.is_under_attack > 1: self.is_under_attack = 1 # у нас не более одного атакующего на здание
        #self.lost_resource()        # теряем производимый ресурс
        self.switch_animation()
# ============================================================
    def reset_under_attack(self):
        self.is_under_attack -= 1
        if self.is_under_attack < 0: self.is_under_attack = 0
        #if self.is_under_attack == 0:
        #    self.start_resource_generation()
        self.switch_animation()
# ============================================================
    def is_broken(self):
        return (self.status == 'broken')
# ============================================================
    def is_main(self):
        return False
# ============================================================
    def on_msg(self, msg_type, msg_param1='', msg_param2='', msg_dict={}):
        '''Предобработчик сообщений. Отделяет системные сообщения от обычных. И вызывает соответстующие методы обработки.'''
        if msg_type == 'refresh_next_res':
            self.update_next_res()
        elif msg_type == 'refresh_next_res_partial':
            self.update_next_res_partial()
        elif msg_type == 'signal_msg':
            self.on_signal(msg_param1)
        #elif msg_type == 'state_machine':
        #    self.on_state_machine(msg_param1, msg_param2, msg_dict)
        elif msg_type == 'click':
            self.reset_cache()
            self.update_next_res()
            self.on_click(msg_param1)
        elif msg_type != 'on_create':
            self.on_message(msg_type, msg_param1, msg_param2, msg_dict)
        elif msg_type == 'on_create':
            self.on_create()
# ============================================================
    def on_msg_now(self, msg_type, msg_param1='', msg_param2='', msg_dict={}):
        if msg_type == 'mouse_in':
            self.on_mouse_in(msg_param1)
        elif msg_type == 'mouse_out':
            self.on_mouse_out()
# ============================================================
    def on_message(self, msg_type, msg_param1, msg_param2, msg_dict):
        '''Обработчик всех обычных сообщений (которые не были обработаны в on_msg)'''
        pass
# ============================================================
    def on_signal(self, signal):
        '''Обработчик сигналов'''
        #if signal == 'offscreen' and self.get('subtype') == 'verf':
        if signal == 'resource_generated':
            self.FSM.HandleEvent('generated_event')
        elif signal == 'update_button_signal':
            _rnd_time = tools.Rand(450, 550)    # чтобы не нагружать всех в один момент времени
            #try:
            self.Signals.Restart('update_button_signal', _rnd_time) # здесь непонятный баг, вроде при апгрейдах зданий...
            #except:
            #    trace('BUG: self.Signals.Restart("update_button_signal")', self)    # затычка, чтобы не падало и чтобы больше узнать про баг
            _is_in_window = self.IsInWindow()
            if _is_in_window:
                if self.get('locind') == BMgr.GetCurMapIndex():
                    self.update_res_enough_state()
        elif signal == 'show_zone':
            # покажем зону здания
            self.show_action_zone = True
            self.is_zone_hlighted = True
            self.Drawer.ShowImages('zone', True)
        elif signal == 'offscreen':
            if self.is_offscreen: return    # чтобы не дублировать
            self.is_offscreen = True
            if self.has_pending_actions:
                self.has_pending_actions = False    # строго до 'try_do_postpoined_action' !
                TMgr.try_do_postpoined_action(self)
        elif signal == 'onscreen':
            self.is_offscreen = False
            self.update_next_res()
        elif signal == 'ah_signal':
            self.anim_helper.start_next_animation()
        elif signal == 'got_access':
            self.refresh_arrow_color()
        elif signal == 'no_access':
            self.Drawer.SetSequence('arrow', 1)    # в любом случае красная будет
            self.switch_animation()
        elif signal == 'building_need_afterdeath':
            self.FSM.HandleEvent('removing_final_event')
        elif signal == 'add_afterdeath_item':
            self.add_afterdeath_item()  # добавляем воронку под зданием
        elif signal == 'remove_countdown':
            # stop round bar
            self.FSM.HandleEvent('removing_event')
        elif signal == 'building_dead':
            self._on_predestroy()
            self.FSM.HandleEvent('removed_event')
        elif signal == 'bonus_restore_eff':
            self.restore_eff_k = 1
        elif signal == 'call_unit_with_res':
            self._call_unit_with_res()
            #self.on_restore_eff_bonus_stop()
# ============================================================
    def refresh_arrow_color(self, redraw=True):
        _seq = 0
        _is_green_state = self.is_green_state()
        if not _is_green_state: _seq = 1
        self.Drawer.SetSequence('arrow', _seq)
        if redraw:
            self.switch_animation()
# ============================================================
    def update_res_enough_state(self, check_green = True):
        '''По сути это уже не только обновление состояния достаточности ресурсов'''
        #@opt попробовать полностью вынести в C++ (а как быть с переопределениями?)
        prev_button_res_enough = self.button_res_enough
        if self.next_res:
            self.button_res_enough = UStats.IsResEnough(self.next_res)#, self.get('uid')) #<< нужно бы IsResEnough проверить по UID и что туда приходит
        else:
            self.button_res_enough = True   #@@ затычка от красного цвета
        if prev_button_res_enough != self.button_res_enough:
            self.switch_animation()

        if check_green:
            self.refresh_arrow_color(False)
# ============================================================
    def on_awaiting_action(self):
        pass
# ============================================================
    def on_functional(self):
        '''Вызывается после ремонта/апгрейда/постройки/активации'''
        pass
# ============================================================
    def on_non_functional(self):
        '''Вызывается после поломки/уничтожения'''
        pass
# ============================================================
    def _on_repair_complete(self, building_info):
        '''Для обработки ситуации окончания repair'''
        BMgr.actions_history.add_action(self.get('locind'), building_info.uid, 'repair_complete')
        if not self.is_deactivated:
            if self.status == 'normal':
                self.on_functional()
        self.on_repair_complete(building_info)
# ============================================================
    def _on_upgrade_complete(self, building_info):
        '''Для обработки ситуации окончания upgrade'''
        BMgr.actions_history.add_action(self.get('locind'), building_info.uid, 'upgrade_complete')
        if not self.is_deactivated:
            if self.status == 'normal':
                self.on_functional()
        self.on_upgrade_complete(building_info)
# ============================================================
    def _on_build_complete(self, building_info):
        '''Для обработки ситуации окончания build'''
        BMgr.actions_history.add_action(self.get('locind'), building_info.uid, 'build_complete')
        #self.on_functional()
        self.on_build_complete(building_info)
# ============================================================
    def _on_first_unit_in(self):
        '''Для обработки ситуации прихода юнита в здание'''
        self.on_functional()
# ============================================================
    def _on_predestroy(self):
        '''For overriding'''
        pass
# ============================================================
    def on_repair_complete(self):
        '''Для обработки ситуации окончания repair'''
        self.send_extra_event()
# ============================================================
    def on_upgrade_complete(self, building_info):
        '''Для обработки ситуации окончания upgrade'''
        self.send_extra_event()
# ============================================================
    def on_build_complete(self, building_info):
        '''Для обработки ситуации окончания build'''
        pass
# ============================================================
    def on_survived_activate(self): #, building_info):
        '''Для обработки ситуации активации выживших зданий'''
        self.on_functional()
# ============================================================
    def _on_broken(self):
        self.on_non_functional()
        self.on_broken()
# ============================================================
    def on_broken(self):
        pass
# ============================================================
    def on_cancel_build(self):
        pass
# ============================================================
    def on_build_started(self):
        pass
# ============================================================
    def on_click(self, btn_name):
        '''Вызывается при клике по объекту'''
        _has_access = self.HasAccess()
        if not _has_access:
            trace('Cancel click (access)', self.get('name') )
            LBoard.CrossNoToLastClick()
            return
        _has_barbarians = UStats.GetLevelValue('barbarian_attack') > 0
        if _has_barbarians:
            return
        _is_active = self.is_active(True)
        _fsm_state = self.FSM.GetStateID()
        if not _is_active:
            _has_remove_or_restore = self.has_remove_or_restore(_fsm_state)
            if _has_remove_or_restore:
                _is_active = True
        if not _is_active:
            trace('Cancel click (inactive)', self.get('name') )
            trace('Cancel info:', self.is_deactivated, self.non_selectable, self.get_next_level_name(), self.is_under_attack)
            #LBoard.MoveCameraToLastClick()
            LBoard.CrossNoToLastClick()
            return
        if _fsm_state in ['wait_unit','removing', 'removing_final']:
            trace('Cancel click ('+_fsm_state+')', self.get('name') )
            LBoard.CrossNoToLastClick()
            return
        self.on_mouse_out() # чтобы обновить состояние селекта
        _has_barbarians = UStats.GetLevelValue('barbarian_attack') > 0
        if self.status in self._broken__place__map_place: _has_barbarians = False
        if self.subtype == 'watchtower': _has_barbarians = False
        #state = self.SM.GetCurrentState()

        if _fsm_state == 'preremoving':
            if btn_name == 'button':
                self.FSM.HandleEvent('cancel_event')
                trace('Cancel click (cancel remove)', self.get('name') )
                trace('Cancel info:', self.is_deactivated, self.non_selectable, self.get_next_level_name(), self.is_under_attack)
            return
        if self.is_upgrade_units_waiting: # потому что ожидание юнита для апгрейда в события запихнуть проблематично
            if _fsm_state != 'upgrading':
                if btn_name == 'button':
                    task = self.get_participating_task('upgrade_building')
                    if task:    # если отменяем апгрейд
                        task.cancel()
                    self.PlaySoundG('SND_CLICK_TASK_CANCEL')
                    self.is_upgrade_units_waiting = False
                    BMgr.UnsuspendPriority(self.get('name'))
                    BMgr.SuspendBuildTemplate(self.get('name'), False)
                    self.is_priority_building = False
                    self.SM.SetState('normal')
                    return
        elif _fsm_state == 'repair_wait':
            if btn_name == 'button':
                self.FSM.HandleEvent('cancel_event')
                return
        elif _fsm_state == 'build_wait':
            if btn_name == 'button':
                _is_fixed_place = self.GetAttrBool('fixed_place')
                if _is_fixed_place:
                    self.FSM.HandleEvent('cancel_mapplace_event')
                else:
                    self.FSM.HandleEvent('cancel_event')
                return
        _buttons_id = self.Get_menu_items_id()
        _buttons_count = len(_buttons_id)
        if _buttons_count > 0:
            self.PlaySoundG('SND_MENU_PLUS')
            self.ShowPopupMenu()
            return


        #_not_enough_workers = BMgr.Is_build_action_possible(map_index, _name, _need_worker)
        #if not _not_enough_workers:
        #    trace('Cancel click (no_worker)', self.get('name') )
        #    #LBoard.MoveCameraToLastClick()
        #    LBoard.CrossNoToLastClick()
        #    return
        #_upgrade_repair_started = self.try_start_upgrade_repair(btn_name)
        #if _upgrade_repair_started:
        #    pass
        #elif state == 'awaiting_action':    # or self.status == "place":
        #    if btn_name != 'button':
        #        LBoard.CrossNoToLastClick()
        #else:
        #    LBoard.CrossNoToLastClick()
        #if not was_remove:
        #    self.on_mouse_in(btn_name)
# ============================================================
    def try_start_upgrade_repair(self, btn_name):
        _name = self.get('name')
        _need_worker = not self.GetAttrBool('need_no_worker')
        map_index = self.get('locind')
        _not_enough_workers = not BMgr.Is_build_action_possible(map_index, _name, _need_worker)
        if _not_enough_workers:
            trace('Cancel click (no_worker)', self.get('name') )
            #LBoard.MoveCameraToLastClick()
            LBoard.CrossNoToLastClick()
            return
        _fsm_state = self.FSM.GetStateID()
        state = self.SM.GetCurrentState()
        _is_next_level_possible = self.is_next_level_possible()
        if _fsm_state not in self._non_cancelable_sates and self.status != 'place':
            if _is_next_level_possible:
                if btn_name in ['button', 'default', 'floor', '_click_rect_']:
                    if self.status == 'normal':
                        self.PlaySoundG('SND_UPGRADE_DONE')
                    else:
                        self.PlaySoundG('SND_CLICK_TASK')
                    if _fsm_state == 'broken':
                        self.FSM.HandleEvent('repair_wait_event')
                    else:
                        self.is_upgrade_units_waiting = True
                        BMgr.SuspendPriority(self.get('name'))
                        BMgr.SuspendBuildTemplate(self.get('name'), True)
                        self.is_priority_building = True
                        self.on_awaiting_action()
                        TMgr.add_task( task_upgrade.UpgradeTask(self) )
                        self.show_arrow(False)
                    return True
            #else:
            #    if self._need_restore_eff_button():
            #        self.restore_eff()
            #    elif self._need_remove_button():
            #        self.FSM.HandleEvent('remove_event')
        else:
            LBoard.CrossNoToLastClick()
            return True
        return False
# ============================================================
# ============================================================
    _no_highlight_states = ['building', 'upgrading', 'repairing', 'map_place']
    def on_mouse_in(self, over_zone):
        if over_zone == 'button':
            return
        if self.status == 'normal':
            if not self.lock_shown:
                self.Signals.Restart('show_zone', 700)
        state = self.SM.GetState()
        _fsm_state = self.FSM.GetStateID()
        #if state == 'awaiting_action':  # ожидаем строителей
        _is_wait_worker_state = self.is_wait_worker_state(_fsm_state)
        self.colorize_zone()
        if _is_wait_worker_state:
            if over_zone != 'button':
                if self.is_mouse_over_zone != 'button':   # если навели/убрали не на/с кнопку
                    self.is_mouse_over = False
                    self.is_selected = False
                    return
        self.is_mouse_over_zone = over_zone
        _is_active = self.is_active()
        if not _is_active:
            _has_remove_or_restore = self.has_remove_or_restore(_fsm_state)
            if _has_remove_or_restore:
                _is_active = True
        if not _is_active:
            self.is_mouse_over = False
            self.is_selected = False
            return
        self.is_mouse_over = True
        _need_highlight = True
        #self.Drawer.HighlightImages(self.highlight_layers, True)
        if _fsm_state in self._no_highlight_states:   # or self.status == 'map_place':
            _need_highlight = False
        self.Drawer.HighlightImages(self.highlight_layers, _need_highlight)
        self.check_button_only_selection(_fsm_state)
        btn_frame = self.Get_button_frame()
        if btn_frame >= 0:
            self.Drawer.SetFrame('button', btn_frame)
        if state in self._normal__empty:
            if not self.is_deactivated:
                if _fsm_state != 'wait_unit':
                    if not self.clock_shown:
                        self.is_selected = True
                        self.show_select(True)
                        return
        self.is_selected = False
        self.show_select(False)
        self.Drawer.HighlightImages(self.highlight_layers, False)
# ============================================================
    def colorize_zone(self):
        _has_color = self.HasAttr('zone_color')
        if _has_color:
            _color = self.GetAttrInt('zone_color')
            self.Drawer.ColorizeImages('zone', _color)
# ============================================================
    def on_mouse_out(self):
        self.is_zone_hlighted = False
        self.show_action_zone = False
        self.Drawer.ShowImages('zone', False)
        self.Signals.Cancel('show_zone')
        self.RemoveMsg('signal_msg', 'show_zone')
        _is_active = self.is_active()
        _fsm_state = self.FSM.GetStateID()
        _has_remove_or_restore = self.has_remove_or_restore(_fsm_state)
        if not _is_active:
            if _has_remove_or_restore:
                _is_active = True
        if not _is_active:
            self.is_mouse_over = False
            self.is_selected = False
            return
        self.is_mouse_over = False
        self.is_mouse_over_zone = ''
        self.Drawer.HighlightImages(self.highlight_layers, False)
        btn_frame = self.Get_button_frame()
        if btn_frame >= 0:
            self.Drawer.SetFrame('button', btn_frame)
        self.is_selected = False
        self.show_select(False)
# ============================================================
    def on_before_building_complete(self):
        '''Вызывается перед апгрейдом здания. Включая начальную инициализацию'''
        self.init_building_cells()
# ============================================================
    def on_upgrade_cancel(self):
        pass
# ============================================================
    def on_upgrade_started(self):
        pass
# ============================================================
    def update_variables(self):
        self.max_units_count = self.GetAttrInt('max_units')
        self.build_progress = 0
        self.build_time = self.GetAttrInt('build_time')
        self.max_build_units = self.GetAttrInt('max_build_units')
        self.res_generation_time = self.GetAttrInt('res_gen_time')
        self.start_cell_pos = self.GetAttrCellFixed('start_cell')
        if self.HasAttr('res_cell'):
            self.res_cell = self.GetAttrCellFixed('res_cell')
        else:
            self.res_cell = self.start_cell_pos # если у здания нет ресурсной клетки, то стартовая подменяет ее
        self.prev_status = self.status
        self.status = self.GetAttrStr('status')
# ============================================================
    def on_building_upgrade(self):
        self.update_variables()
        _is_home = self.is_home()
        if _is_home: self.SetIsHome(True)
        self.add_fog_circle()
        self.add_water_circle()
        self.init_default_images()
        self.init_building_cells()
        # если нужно, то создадим юнитов внутри здания:
        self.create_building_units(self.external_units_count)
        #@fsm:
        #if not self.is_deactivated:
        #    self.call_units_or_start_resgen()
        self.update_next_res()
        if self.is_shaded:
            self.Drawer.SetShaded(self.shaded_layers, True)
        self.non_selectable = self.GetAttrBool('non_selectable')
        if self.non_selectable:
            self.SetMouseActive(False)
        self.switch_animation()
        self.SendMsg('upgrade_complete')
# ============================================================
    def on_building_create(self):
        '''Метод инициализации здания при смене варианта или первичном размещении на карте
            Вырезает туман войны, строит дороги и прочее...'''
        self.init_default_images()
        self.update_variables()
        _is_home = self.is_home()
        if _is_home:
            self.SetIsHome(True)
        self.add_fog_circle()
        self.add_water_circle()
        # если нужно, то создадим юнитов внутри здания:
        self.external_units_count = self.GetAttrInt('external_units')
        self.create_building_units(0)
        self.update_next_res()
        #if not self.is_deactivated:
        #    self.call_units_or_start_resgen()
        # ###
        self.reinit_effectiveness_params() # после self.on_building_create()  (потому что res_generation_time)
        _rnd_time = tools.Rand(450, 550)
        self.Signals.Restart('update_button_signal', _rnd_time)   # обновляем состояние кнопки раз в полсекунды
        self.refresh_arrow_color()
        if self.status == 'normal':
            if not self.is_deactivated:
                self.FSM.HandleEvent('prenormal_event')
            else:
                self.FSM.HandleEvent('survived_event')
        elif self.status == 'broken':
            _vname = self.get('variant_name')
            if 'thrown' in _vname:
                self.FSM.HandleEvent('thrown_event')
            else:
                self.FSM.HandleEvent('broken_event')
        elif self.status == 'map_place':
            self.FSM.HandleEvent('map_place_event')
# ============================================================
    def need_call_units(self):  #@fsm
        _is_home = self.is_home()
        if _is_home:
            return False
        if self.is_indians:
            return False
        if self.GetAttrBool('need_no_worker'):
            return False
        if not self.worker_units:       # если юнитов ранее еще не вызывали
            return True
        else:
            return False
# ============================================================
    def call_units(self):   #@fsm
        _need_call_units = self.need_call_units()
        if _need_call_units:
            TMgr.add_task( task_callworker.CallWorkerTask(self) )   # позовем юнита для начала работ
# ============================================================
    def add_water_circle(self):
        '''Добавляет зону доступности воды для зданий'''
        if self.status != 'broken':
            _stype = self.get('stype')
            if _stype in ['well', 'waterpump']: # только эти здания
                self.AddWaterCircle(1)
# ============================================================
    def add_fog_circle(self):
        '''Данный метод нужен для того, чтобы его можно было переопределять'''
        if self.status != 'broken':
            trace('add_fog_circle:', self.prev_status, self.status)
            self.AddFogCircle(self.is_main())
            if self.subtype in ['totem', 'totem_stone']:
                if self.prev_status not in ['__none__', '']:
                    if self.prev_status == 'normal' and self.status == 'normal':
                        pass
                    else:
                        trace('TOTEM:', self.prev_status, self.status)
                        self.AddEffect('fog_off_big', self.get('position'))
                        self.PlaySoundG('SND_NEW_AREA')
# ============================================================
    def on_create(self):
        '''Метод инициализации при первичной установке на карту (не важно предустановка это или постройка)'''
        self.init_fsm()
        self.on_building_create()  # вызывать только здесь, чтобы все объекты уже установились на карту
# #######################################################
# #######################################################
# #######################################################
# ============================================================
    def is_builders_state(self):
        _fsm_state = self.FSM.GetStateID()
        if _fsm_state in self.builders_states:
            return True
        if self.is_upgrade_units_waiting:
            return True
        return False
# ============================================================
    def is_wait_worker_state(self, fsm_state):
        #_fsm_state = self.FSM.GetStateID()
        if fsm_state in self.wait_states:
            return True
        if self.is_upgrade_units_waiting:
            return True
        return False
# ============================================================
    _floor__icons = ['floor','icons']
    _default__anim = ['default','anim']
    _all_remove_states = ['preremoving','removing','removing_final']
    _real_remove_states = ['removing','removing_final']
    def switch_animation(self):
        _fsm_state = self.FSM.GetStateID()
        state = self.SM.GetCurrentState()
        drawer = self.Drawer
        # далее финальная часть, чтобы в тумане не вылазили части объекта:
        if drawer.IsPartiallyInFog():
            drawer.ShowImages('', False)
            drawer.ShowImages('floor')
            drawer.SetSequence('floor', self.current_episode-1)
            return
        drawer.ShowImages('', False)
        drawer.ShowImages(self._floor__icons)
        drawer.SetSequence('floor', self.current_episode-1)
        drawer.ShowImages('resources', self.is_generated_res)
        self._switch_animation_extra('begin')
        if state in self._normal__empty:         # '' (empty) - это состояние после выхода из любого другого
            drawer.ShowImages(self._default__anim)
            drawer.ShowImages('unit', self.need_to_show_unit())
            drawer.ShowImages('no_action', self.need_no_action())
            drawer.ShowImages('pers_no_action', self.need_pers_no_action())
            has_worker = len(self.worker_units) > 0
            if self.is_under_attack == 0:
                drawer.ShowImages('worker_inside', has_worker)
            _need_no_unit = not has_worker and self.status == 'normal'
            drawer.ShowImages('no_unit', _need_no_unit)
            self._switch_animation_extra('normal')
        #elif state == 'awaiting_action':        # ожидания действия юнита - нужно галку показывать
        elif self.is_wait_worker_state(_fsm_state):
            drawer.ShowImages(['default','anim'])
            drawer.ShowImages('unit', self.need_to_show_unit())
            drawer.ShowImages('no_action', self.need_no_action())
            self._switch_animation_extra('awaiting')
        elif _fsm_state == 'building':
            drawer.ShowImages('build')
            drawer.SetProgressedAnimation('build')
            #self.Drawer.RestartEffect('build')
            self._switch_animation_extra('build')
        elif state == 'upgrade':
        #elif _fsm_state in ['upgrading', 'repairing']: # не прокатывает, т.к. в b_boatyard 'UPGRADE' выставляется как-то хитро
            #if self.get('subtype') not in ['verf', 'boatyard', 'dam']:
            if not self.IsSubTypeOf(['verf', 'boatyard', 'dam']):
                if self.status == 'broken':
                    drawer.ShowImages('default')  # у сломанного состояния нет доп. изображения
                    drawer.ShowImages('repair')
                    #self.Drawer.RestartEffect('repair')
                else:
                    drawer.ShowImages('upgrading')
                    #self.Drawer.RestartEffect('upgrading')
            #if self.status not in self._place__broken:
            #    self.Drawer.ShowImages("anim")      # анимации можно показывать и при строительстве, хотя, возможно, в некоторых зданиях придется их отключать
            self._switch_animation_extra('upgrade')
        btn_frame = self.Get_button_frame()
        if btn_frame >= 0:  drawer.SetFrame('button', btn_frame)
        self.show_select(not self.is_deactivated and self.is_selected)
        _is_active = self.is_active()
        _has_remove_or_restore = self.has_remove_or_restore(_fsm_state)
        if not _is_active:
            if _has_remove_or_restore:
                _is_active = True
        if not _is_active or self.is_builders_state():   # state in ['build', 'upgrade', 'awaiting_action']:
            self.show_select(False)
            drawer.HighlightImages(self.highlight_layers, False)
        if _fsm_state == 'wait_unit':
            self.show_select(False)
            drawer.HighlightImages(self.highlight_layers, False)
        self.check_button_only_selection(_fsm_state)
        drawer.ShowImages('arrow', self.arrow_shown)
        drawer.ShowImages('clock_wait', self.clock_shown)
        drawer.ShowImages('lock', self.lock_shown)
        if not self.is_builders_state(): #state not in ['build', 'upgrade', 'awaiting_action']:
            if self.show_action_zone:
                drawer.ShowImages('zone', self.is_zone_hlighted)  # для b_woodcutter нужно
            else:
                drawer.ShowImages('zone', False)
        else:
            drawer.ShowImages('zone', False)
        self.anim_helper.show_last_anim()
        # ---
        if _fsm_state in self._all_remove_states:
            if _fsm_state in self._real_remove_states:
                drawer.ShowImages('boom')
            drawer.ShowImages(['anim','pers_no_action','unit','worker_inside', 'sleep_unit', 'resources'], False)
            drawer.StopEffect('anim')
        # ---

        self._switch_animation_extra('end')     # всегда в конце
# ============================================================
    def check_button_only_selection(self, _fsm_state):
        _is_cancel_button_mode = self.is_cancel_button_mode(_fsm_state)
        if _is_cancel_button_mode:
            if self.is_mouse_over_zone != 'button':
                self.Drawer.HighlightImages(self.highlight_layers, False)
                self.show_select(False)
                self.is_selected = False
            else:
                self.Drawer.HighlightImages(self.highlight_layers, False)
                self.show_select(False)
                self.is_selected = False
                    #    self.Drawer.ShowImages('select')
# ============================================================
    def _switch_animation_extra(self, place):
        '''Метод для переопределения в подклассах
        place: точка вызова метода'''
        pass
# ============================================================
    def Invalidate(self):
        '''перерисовка объекта, вызывается из C++'''
        self.switch_animation()
# ============================================================
    def need_to_show_unit(self):
        '''Нужно ли показывать юнита в здании'''
        return self.is_generating_res and self.is_under_attack == 0
# ============================================================
    def need_no_action(self):
        '''Нужно ли показывать юнита в здании'''
        return not self.is_generating_res
# ============================================================
    def need_pers_no_action(self):
        '''Нужно ли показывать юнита в здании'''
        return not self.is_generating_res and self.worker_units and self.is_under_attack == 0
# ============================================================
    def reinit_effectiveness_params(self, force = False):
        '''Переинициализирует параметры эффективности здания'''
        _can_reinit = False
        if not self.has_effectivness_inited: _can_reinit = True
        if force:
            _can_reinit = True
            self.res_generation_time = self.GetAttrInt('res_gen_time')
        if _can_reinit:
            self.need_max_time = False
            self.has_effectivness_inited = True
            _new_max_res_generation_time = self.GetAttrInt('res_gen_time_max')
            _new_max_res_gen_time = self.GetAttrInt('gen_max_time')
            if not force:
                if _new_max_res_generation_time == self.max_res_generation_time:
                    if _new_max_res_gen_time == self.max_res_gen_time:
                        return  # если параметры не изменились, то значит значения сохраняем
            self.res_gen_first_times = self.GetAttrInt('res_gen_first_times')
            self.max_res_gen_time = self.GetAttrInt('res_gen_maxtime')
            _need_init_gen_first_time = False
            if self.current_effectiveness == 100:
                if self.res_gen_first_times > 0:  # если первый отсчет еще
                    if not force:   # только в первом случае, после восстановления эффективности этого делать не нужно
                        _need_init_gen_first_time = True
            if _need_init_gen_first_time:
                self.effectiveness_degradation_count = self.res_gen_first_times
            else:
                self.effectiveness_degradation_count = self.GetAttrInt('res_gen_ef_count')   # через сколько генераций ресурсов эффективность падает на 10%
            if self.effectiveness_degradation_count == 0:
                self.effectiveness_degradation_count = 3    # дефолтное значение
            self.max_res_generation_time = self.GetAttrInt('res_gen_time_max')
            if self.max_res_generation_time < self.res_generation_time:
                self.max_res_generation_time = self.res_generation_time + 1 # +1 чтобы избежать деления на ноль
            self.generation_count = 0               # сколько раз уже генерировали ресурс
            _prev_eff = self.current_effectiveness
            self.current_effectiveness = 100        # текущая эффективность (проценты)
            if self.current_effectiveness != _prev_eff:
                EMgr.SendGlobalMsg('effectiveness_changed', '')
# ============================================================
    def update_effectiveness(self):
        '''Нужно вызывать при генерации ресурса зданием'''
        if self.Signals.Has('bonus_restore_eff'):   # если бонус, то сразу выходим и ничего не делаем
            return
        else:
            self.restore_eff_k = 1.0    # для restore_eff сбросим коэффициент в единицу
        self.generation_count += 1
        if not self.need_max_time:
            if self.current_effectiveness == 10:    # если это последний отсчет
                self.need_max_time = True
                self.on_minimal_effectiveness()
        if self.effectiveness_degradation_count:
            if self.generation_count >= self.effectiveness_degradation_count:
                self.effectiveness_degradation_count = self.get_attr_int('res_gen_ef_count')  # потому что первая генерация отличается от остальных
                self.generation_count = 0
                if self.current_effectiveness > 10:     # 10% - минимальная эффективность
                    _prev_eff = self.current_effectiveness
                    self.current_effectiveness -= 10    # снимаем 10%
                    #_need_help = LBoard.Core.is_help_needed(4)
                    #if _need_help:
                    #    _item = BMgr.get_item('2542', self.get('locind'))
                    #    if _item: Camera.MoveTo(_item)
                    #    UI.open_help(True)
                    if LBoard.Core.episode > 1 or LBoard.Core.level > 2:    # начиная с 1_3 уровня
                    #if True:
                        _notify_started = UStats.GetLevelValue('effectiveness_notify_started')
                        if not _notify_started:
                            _is_blocked = UStats.GetLevelValue('BLOCK_RESTORE_EFF') > 0
                            if not _is_blocked:
                                #if _prev_eff >= 50: # для подсказок - они должны стартовать после падения эффективности менее 80% любого здания
                                    if self.current_effectiveness < 50:
                                        EMgr.SendGlobalMsg('effectiveness_50', '')
                                        UStats.AddLevelValue('effectiveness_notify_started')
                    if self.current_effectiveness < 10: self.current_effectiveness = 10
                    if self.current_effectiveness != _prev_eff:
                        EMgr.SendGlobalMsg('effectiveness_changed', '')
# ============================================================
    def get_attr_int(self, attr_name):
        '''Чтобы можно было легко переопределить в полях'''
        return self.GetAttrInt(attr_name)
# ============================================================
    def on_minimal_effectiveness(self):
        pass
# ============================================================
    def get_effectiveness_coeff(self, skip_bonus = False):
        '''Возвращает текущую эффективность здания в диапазоне [0.0 .. 1.0]'''
        _has_bonus = self.Signals.Has('bonus_restore_eff')
        if skip_bonus: _has_bonus = False
        if _has_bonus:
            return 1.0
        else:
            return (self.current_effectiveness / 100)
# ============================================================
    def get_current_lowest_effectiveness(self):
        '''Для возможности переопределения в случае с фермой'''
        return self.current_effectiveness
# ============================================================
    def get_effective_gen_time(self, effectiveness = -1):
        '''Возвращает время генерации ресурса с учетом эффективности здания'''
        if self.subtype in self.has_no_effectiveness:
            return self.res_generation_time
        if effectiveness == -1:
            if self.need_max_time:
                if self.max_res_gen_time:
                    return self.max_res_gen_time    # если требуется максимальное время генерации
        if effectiveness == -1: effectiveness = self.current_effectiveness
        #coeff = self.get_effectiveness_coeff()
        coeff = (effectiveness / 100)
        gen_time = self.res_generation_time
        max_gen_time = self.max_res_generation_time
        return gen_time + ((1 - coeff)/0.9)*(max_gen_time - gen_time)
# ============================================================
    def switch_level(self, action = '_none_', extra_text = ''):
        '''Здесь происходит "апгрейд" здания, смена варианта. Здесь должны (!) проверяться все необходимые требования, кроме кол-ва ресурсов'''
        var_name = self.get('variant_name')    # текущий вариант
        if self.max_variant == var_name:
            return      # уже максимальный уровень
        if var_name in self.variants_switch:
            _prev_status = self.status
            next_var_name = self.variants_switch[var_name]
            self.show_arrow(False)
            self.energy = 100   # на всякий случай восстановим энергию
            trace('switch_level: action =', action)
            trace(' switch_level:', self.get('name'), self.get('uid'), next_var_name)
            self.reset_cache()
            self.SwitchLevel(next_var_name)   # переключаемся на следующий вариант
            self.energy_hit = self.GetAttrInt('action_hit') # чтобы при смене вариантов обновлялось
            self.update_next_res()
            self.switch_animation()
            self.unreserve_action_cells()
            #if _prev_status == 'broken':    # если здание было сломано, то
            # отремонтировано обычное здание, либо отремонтировано/построено priority-здание
            if next_var_name != 'place':
                BMgr.AddRepaired( self.get('name'), next_var_name )    # добавим его в список отремонтированных
                BMgr.RemoveBuildTemplate(self.get('name'), True)
            _stype = self.get('stype')
            if _stype not in ['farm_field', 'forester_field']:
                self.AddEffect('build_end', self.get('position'))
                self.PlaySoundG('SND_BUILDING_COMPLETE')
            letter = []
            _skip_stype = self.IsSubTypeOf(['farm_field', 'forester_field', 'fence']) # эти не должны ничего подобного генерировать
            if not _skip_stype:
                if action:
                    if action == 'builded': letter.append( ('text', 'MSG_BUILDED') )
                    elif action == 'upgraded': letter.append( ('text', 'MSG_UPGRADED') )
                    elif action == 'repaired': letter.append( ('text', 'MSG_REPAIRED') )
                    elif action == 'boat_build': letter.append( ('text', 'MSG_BOAT_COMPLETE') )
                    letter.append( 'space' )
                    letter_item_name = self.GetAttrStr('letter_name')
                    if not letter_item_name:    # если значение не определено
                        letter_item_name = self.GetAttrStr('item_name')
                    letter.append( ('text', letter_item_name) )
                else:
                    letter = extra_text
                _loc_ind = self.get('locind')
                EMgr.AppendLetter(letter, self.get('uid'), _loc_ind)
                #if action in ['builded', 'upgraded']:
                #    self.Drawer.ShowImages('star_effect')
                #    self.Drawer.RestartEffect('star_effect')
            _non_selectable = self.GetAttrBool('non_selectable')
            if _non_selectable:             # может меняться только в сторону невыделяемости, но не обратно
                self.SetMouseActive(False)
                self.non_selectable = True
                self.is_mouse_over = False
                self.is_selected = False
                self.Drawer.ShowImages('select', False)
                self.Drawer.HighlightImages(self.highlight_layers, False)
            self.on_switch_level_complete()
        self.update_res_enough_state()      #@? возможно лишний вызов, в update_next_res должно вызываться
# ============================================================
    def on_switch_level_complete(self):
        self.boost_work_percent = self.GetAttrInt('boost_work_percent')
# ============================================================
    def switch_to_blocked_level(self):
        '''Здесь происходит "апгрейд" здания, смена варианта. Здесь должны (!) проверяться все необходимые требования, кроме кол-ва ресурсов'''
        var_name = self.get('variant_name')    # текущий вариант
        if var_name in self.blocked_switch:
            _prev_status = self.status
            next_var_name = self.blocked_switch[var_name]
            self.show_arrow(False)
            trace('switch_to_blocked_level:')
            trace(' switch_to_blocked_level:', self.get('name'), self.get('uid'), next_var_name)
            self.reset_cache()
            self.SwitchLevel(next_var_name)   # переключаемся на следующий вариант
            self.update_next_res()
            self.switch_animation()
            self.unreserve_action_cells()
            self.is_thrown = True
            self.on_switch_level_complete()
        self.update_res_enough_state()  #@? лишнее?
# ============================================================
    def switch_level_to_place(self):
        '''Здесь происходит смена шаблона с map_place на place'''
        _is_fixed_place = self.GetAttrBool('fixed_place')
        if _is_fixed_place:     # только для тех зданий, где фиксированное место
            var_name = self.get('variant_name')    # текущий вариант
            #_mplace = 'map_place'
            #_var_name_part = var_name[:len(_mplace)]
            _is_map_place = var_name.startswith('map_place')
            if _is_map_place:
                self.prev_mplace_name = var_name
                next_var_name = self.variants_switch[var_name]
                self.show_arrow(False)
                trace('switch_level_to_place:')
                trace(' switch_level_to_place:', self.get('name'), self.get('uid'), next_var_name)
                self.reset_cache()
                self.SwitchLevel(next_var_name)   # переключаемся на следующий вариант
                self.update_next_res()
                self.switch_animation()
                self.on_switch_level_complete()
# ============================================================
    def cancel_switch_level_to_place(self):
        '''Отмена смены шаблона с map_place на place, т.е. переключение наоборот: place на map_place'''
        _is_fixed_place = self.GetAttrBool('fixed_place')
        if _is_fixed_place:     # только для тех зданий, где фиксированное место
            var_name = self.get("variant_name")    # текущий вариант
            if var_name == 'place':
                next_var_name = self.prev_mplace_name   #self.cancel_switch_place_variant_name
                self.show_arrow(False)
                trace('cancel_switch_level_to_place:')
                trace(' cancel_switch_level_to_place:', self.get('name'), self.get('uid'), next_var_name)
                self.reset_cache()
                self.SwitchLevel(next_var_name)   # переключаемся на предыдущий вариант
                self.update_next_res()
                self.switch_animation()
                self.unreserve_action_cells()
                self.on_switch_level_complete()
# ============================================================
    def update_next_res(self):
        '''Обновляет значение ресурсов, необходимых для апгрейда на следующий уровень'''
        #_stype = self.get('stype')
        _is_home = self.is_home()
        _need_worker = not self.GetAttrBool('need_no_worker')
        if _need_worker and self.status == 'normal' and not self.worker_units and not _is_home:
            self.has_next_level = False
        else:
            next_var_name = self.get_next_level_name()
            if next_var_name:
                self.has_next_level = True
                self.next_res = self.get_needs_res(next_var_name)
            else:
                self.has_next_level = False
        self.update_res_enough_state()
        btn_frame = self.Get_button_frame()
        if btn_frame >= 0:
            self.Drawer.SetFrame('button', btn_frame)
# ============================================================
    def update_next_res_partial(self):
        '''Обновляет значение ресурсов, необходимых для апгрейда на следующий уровень'''
        #_stype = self.get('stype')
        _is_home = self.is_home()
        _need_worker = not self.GetAttrBool('need_no_worker')
        if _need_worker and self.status == 'normal' and not self.worker_units and not _is_home:
            self.has_next_level = False
        else:
            next_var_name = self.get_next_level_name()
            if next_var_name:
                self.has_next_level = True
                self.next_res = self.get_needs_res(next_var_name)
            else:
                self.has_next_level = False
# ============================================================
    def create_building_units(self, external_units_count):
        '''Создает необходимое кол-во юнитов в здании, если у здания они должны быть
            При апгрейде здания досоздает необходимое кол-во'''
        cur_units_count = self.GetAttrInt('start_units')
        if self.status not in self._place__broken:
            for i in range(self.cur_units_count + external_units_count, cur_units_count):      # добавим недостающее кол-во юнитов
                unit = UMgr.create_unit(self.get('locind'), self.action_unit_name)
                self._building_units.append(unit)
                unit.set_home(self)
                unit.CheckIsInFog(True)     # проверка на то, что юнит не в тумане и прочее
                if self.is_internal_free_units:
                    self._internal_free_units.append(unit)
                else:
                    UMgr.add_free_unit(unit)     # новые юниты свободны: у нас свободное общество!
                EMgr.send_event( '_worker_added', self )
            self.cur_units_count = cur_units_count # обновим текущее кол-во юнитов(созанных данным зданием)
# ============================================================
    def init_building_cells(self):
        # сделаем выборку ячеек (имен) строительства (от 1 до 8 точек):
        del self.building_cells_names[:]
        for i in range(1,8+1):
            cell_name = 'build' + str(i) + '_cell'
            if self.HasAttr(cell_name):
                self.building_cells_names.append(cell_name)
        self.barb_cells_names = []
        for i in range(0,4):
            cell_name = 'attack' + str(i+1) + '_cell'
            if self.HasAttr(cell_name):
                self.barb_cells_names.append(cell_name)
# ============================================================
    def init_default_images(self):
        drawer = self.Drawer
        drawer.ShowImages('', False)
        drawer.ShowImages('floor')
        drawer.ShowImages('default')
        drawer.ShowImages('unit')
        drawer.ShowImages('anim')
        drawer.ShowImages('resources')
# ============================================================
    def _is_base_res_enough(self):   #@fsm
        resource_enough = False
        _need_base_resource = self.need_base_resource()
        if _need_base_resource:
            if self.has_base_res:
                resource_enough = True
        else:
            resource_enough = True
        return resource_enough
# ============================================================
    def _can_generate_resource(self):   #@fsm
        if not self.res_generation_time:
            return False
        if self.is_generated_res:   # 2012-12-19 если ресурс еще не забрали, то генерировать не можем же?
            return False
        res = self.GetRes('normal_gives_1')
        _is_res_empty = True
        if res: _is_res_empty = res.Empty()
        if _is_res_empty:     # если ресурс не задан, то здесь делать нечего
            return False
        res_name = res.GetSingleName()
        if not res_name:
            return False
        return True
# ============================================================
    def _start_resource_generation(self):    #@fsm
        self.is_wait_res_unit = False
        if not self.Signals.Has('bonus_restore_eff'):   # если бонус не запущен, то
            self.restore_eff_k = 1.0    # для restore_eff сбросим коэффициент в единицу???
        self.is_generating_res = True
        self.on_res_gen_started()
        self.Drawer.ShowIn('unit')
        self.Signals.Restart('resource_generated', self.get_effective_gen_time())
        _stype = self.get('stype')
        if _stype not in self.fake_buildings:
            if self.Signals.Has('bonus_restore_eff'):
                self.renew_restore_bonus_eff()
        self.NeedRenewGenerationBoost() # чтобы обновить коэффициент ускорения всеми методами
        #        k = self.get_effectiveness_coeff() / (self.current_effectiveness/100)
        #        self.Signals.SetTimeMul('resource_generated', k)
        self.switch_animation()
# ============================================================
    def on_res_gen_started(self):
        pass
# ============================================================
    def stop_resource_generation(self):
        self.Signals.Cancel('resource_generated')
        self.is_generating_res = False
        self.switch_animation()
# ============================================================
    def generate_resource(self):
        '''Генерирует ресурс возле здания'''
        self.update_effectiveness()
        if not self.is_generated_res:   # если ресурс еще не сгенерирован
            self.is_generating_res = False
            self.is_generated_res = True
            self.Drawer.ShowOut('unit')
            self.Drawer.ShowImages('resources', True)   # показать ресурс
            _pos = self.GetAttrPoint('center_res_pos')
            self.Drawer.ResIn('resources', _pos, 1000)
            self.call_take_res_unit()
            self.is_generated_res_doubled = self.Signals.Has('bonus_gen_double')
            #self.is_generated_res_doubled = UStats.GetLevelValue('BONUS_GENERATE_DOUBLE') > 0
            # - нужно попробовать повторить данный таск, только главное не создать багов при этом
            # например дублей...
        self.switch_animation()
# ============================================================
    def call_take_res_unit(self):
        if self.is_generated_res:
            _is_participating = self.is_task_participating('take_resfrom')
            if not _is_participating:
                cell = self.GetAttrCellFixed('res_cell')
                self.is_wait_take_res_unit = True
                TMgr.add_task( task_res.TakeResTask(self, cell) )   # создать таск "забрать ресурс"
# ============================================================
    def scale_bump(self):
        _pos = self.GetAttrPoint('center_item_pos')
        self.Drawer.ScaleBump(self.bump_layers, _pos, 120, 1.016)
        _is_stone = self.GetAttrBool('is_stone')
        if _is_stone:
            self.PlaySound('SND_HAMMER_HIT_ROCK')
        else:
            self.PlaySoundRandom(['SND_HAMMER_HIT_1', 'SND_HAMMER_HIT_2', 'SND_HAMMER_HIT_3'])
# ============================================================
    def _call_unit_with_res(self):  #@fsm
        _res_name = self.GetSingleResName('base_resource')
        if not _res_name:
            return
        trace('Check - building name:', self.get('name'), _res_name)
        _cell = self.FixAddCell(self.cell_pos, self.GetAttrCell('res_cell'))
        _task = task_res_store.TakeResStoreTask(self, _cell, _res_name)
        TMgr.add_task( _task )  # добавим таск, там установится UID
# ============================================================
    def _restask_self_cancel(self):
        self.is_wait_res_unit = False
        #self.Signals.Restart('call_unit_with_res', 500)    # bug431
# ============================================================
    def move_global_res_to_unit(self, _unit, _res_name, _need_res_value):
        '''Вообще-то данный метод должен быть не в Building, но увы...
        Метод переносит часть глобального ресурса на юнита'''
        _cur_res = UStats.GetCurRes()
        _res_count = _cur_res.Get(_res_name)
        if not _res_count:
            _unit.resource = None
        else:
            if _need_res_value >= _res_count:
                _new_res = UStats.CreateRes(_res_name, _res_count)
            else:
                _new_res = UStats.CreateRes(_res_name, _need_res_value)
            _unit.resource = _new_res
            UStats.SubRes(_new_res)
# ============================================================
    def give_resource(self):
        '''Вызывается в момент, когда юнит забирает ресурс у здания. Скрывает картинку ресурса и начинает новую цепочку производства'''
        self.Drawer.ShowImages('resources', False)
        self.is_generated_res = False # отдаем ресурс
        resource = self.GetRes('normal_gives_1')
        if self.is_generated_res_doubled:
            resource = resource.Clone()
            resource.MultiplyValues(2)
        self.is_wait_take_res_unit = False
        self.is_generated_res_doubled = False
        self.has_base_res = False
        TMgr.send_event('res_taken', None)
        self.SendMsg('resource_was_taken')
        self.FSM.HandleEvent('res_taken_event')
        return resource
# ============================================================
    def lost_resource(self):
        '''Уменьшает на 1 кол-во ресурса у здания, при этом выставляет
        некоторые флаги того, что ресурс отдан
        Используется при атаке дикарями, чтобы "потерять" производимый ресурс'''
        if self.is_generating_res:      # если идет производство ресурса:
            self.is_generated_res = False
            self.is_generating_res = False
            self.has_base_res = False
            self.update_effectiveness()
            self.is_wait_take_res_unit = False
        self.stop_resource_generation()
# ============================================================
    def take_resource(self, resource):
        '''Вызывается в момент, когда юнит отдает ресурс зданию. Может инициировать процесс генерации зависимого ресурса, либо добавляет ресурс к общим ресурсам'''
        #if self.need_base_resource():
        #    self.base_resource_count += resource.GetSingleValue()
        self.has_base_res = True
# ============================================================
    def has_generating_res_progress(self):
        '''Генерирует ли данное здание ресурс в данный момент, или ресурс сгенерирован (100% progress).
            Может ли данное здание принимать ресурсы. Оно не может принимать ресурсы, если
            был принят ресурс и идет процесс производства, либо ресурс сгенерирован, но еще не забран из здания,
            либо есть юнит, который намеревается нести что-то в здание.
            Здесь НЕ должно быть зависимости от типа здания (может ли оно вообще принимать ресурс). Данный метод
            используется при выборе доступных зданий'''
        if self.is_generating_res:
            return True
        if self.is_generated_res:
            return True
        else:
            return False
# ============================================================
    def apply_bonus_double_generation(self, time):
        _stype = self.get('stype')
        if _stype not in self.fake_buildings:
            _can_double_produce = self.GetAttrBool('can_double_produce')
            if _can_double_produce:
                self.Signals.Restart('bonus_gen_double', time)
                #self.Drawer.StartEffectOnce('bonus_double')
                self.start_bonus_effect_double_generation()
# ============================================================
    def start_bonus_effect_double_generation(self):
        self.Drawer.StartEffectOnce('bonus_double')
# ============================================================
    def apply_bonus_restore_effectiveness(self, time):
        _stype = self.get('stype')
        if _stype not in self.fake_buildings:
            _can_restore_bonus = self.GetAttrBool('can_restore_bonus')
            if _can_restore_bonus:
                self.Signals.Restart('bonus_restore_eff', time)
                if self.get_effective_gen_time(100) != 0:
                    self.restore_eff_k = self.get_effective_gen_time() / self.get_effective_gen_time(100)
                else:
                    self.restore_eff_k = 1
                #self.Drawer.StartEffectOnce('bonus_restore')
                self.start_bonus_effect_restore_effectiveness()
# ============================================================
    def start_bonus_effect_restore_effectiveness(self):
        self.Drawer.StartEffectOnce('bonus_restore')
# ============================================================
    def on_restore_eff_bonus_start(self, k):
        self.Signals.SetTimeMul('resource_generated', k)
# ============================================================
    def on_restore_eff_bonus_stop(self):
        self.Signals.SetTimeMul('resource_generated', 1)    # как только закончился бонус - сбросим
# ============================================================
    def renew_restore_bonus_eff(self):
        self.on_restore_eff_bonus_stop()
        self.on_restore_eff_bonus_start(self.restore_eff_k)
# ============================================================
    def is_next_level_possible(self, skipResEnough = False):
        '''Возможен ли апгрейд/ремонт до следующего уровня'''
        next_var_name = self.get_next_level_name()
        if next_var_name:
            _need_worker = not self.GetAttrBool('need_no_worker')
            if _need_worker:
                _is_home = self.is_home()
                if not _is_home:
                    if not self.worker_units:
                        if self.status == 'normal':
                            return False
            res_need = self.get_needs_res(next_var_name)
            _state = self.SM.GetCurrentState()
            if _state in self._normal__empty:   # если не в состоянии стройки и т.д.
                if self.status in self._normal__broken:         # если есть что апгрейдить
                    if res_need:
                        _is_res_enough = True
                        if not skipResEnough:
                            _is_res_enough = UStats.IsResEnough(res_need)
                        if _is_res_enough:          # если ресурсов достаточно
                            return True
        return False
# ============================================================
    def get_next_level_name(self, skipCanRepairCheck = False):  #@opt! Performance hungry
        '''Возвращает имя следующего варианта после текущего'''
        if skipCanRepairCheck:
            if self.cached_next_level_name_skipcheck == '__reset__':
                self.cached_next_level_name_skipcheck = self._get_next_level_name_int(skipCanRepairCheck)
            return self.cached_next_level_name_skipcheck
        else:
            if self.cached_next_level_name_check == '__reset__':
                self.cached_next_level_name_check = self._get_next_level_name_int(skipCanRepairCheck)
            return self.cached_next_level_name_check
# ============================================================
    def _get_next_level_name_int(self, skipCanRepairCheck = False):  #@opt! Performance hungry
        '''Возвращает имя следующего варианта после текущего'''
        _template_name = self.get('name')
        var_name = self.get('variant_name')    # текущий вариант
        map_index = self.get('locind')
        if not skipCanRepairCheck:
            _can_be_repaired = BMgr.CanBeRepairedPriority(_template_name, map_index)
            if not _can_be_repaired:
                return None
        _max_global_var_name = BMgr.GetMaxBuildingVariant(_template_name, map_index)
        _max_global_var_allowed = BMgr.CanBuildMaxVariant(_template_name, map_index)
        if _max_global_var_name:
            if var_name in self.variants_switch:
                _max_index = self.variants_indexes[_max_global_var_name]
                _cur_index = self.variants_indexes[var_name]
                if _cur_index >= _max_index:
                    return None
        if var_name != self.max_variant:        # если максимальный вариант (не глобальный), то апгрейдить уже некуда
            if var_name in self.variants_switch:
                next_var_name = self.variants_switch[var_name]
                if _max_global_var_name:
                    if next_var_name == _max_global_var_name:
                        if not _max_global_var_allowed:
                            if self.status == 'normal': # если максимальный вариант строить еще нельзя
                                return None
                if self.HasVariant(next_var_name):
                    return next_var_name
        return None
# ============================================================
    def is_priority_blocked(self):
        '''Возвращает имя следующего варианта после текущего'''
        _template_name = self.get('name')
        map_index = self.get('locind')
        _can_be_repaired = BMgr.CanBeRepairedPriority(_template_name, map_index)
        if not _can_be_repaired:
            return True
        return False
# ============================================================
    def show_arrow(self, show = True):
        '''Показать/скрыть квестовую стрелку'''
        self.arrow_shown = show
        self.switch_animation()
# ============================================================
    def show_clock(self, show = True):
        '''Показать/скрыть песочные часы'''
        self.clock_shown = show
        self.switch_animation()
# ============================================================
    def show_lock(self, show = True):
        '''Показать/скрыть песочные часы'''
        self.lock_shown = show
        self.switch_animation()
# ============================================================
    def show_effect(self, group):
        '''Показать/скрыть эффект'''
        #self.arrow_shown = show
        self.Drawer.ShowImages(group)
        self.Drawer.StartEffectOnce(group)
# ============================================================
    def __str__(self):
        result = to_str('{ name:', self.get('name'), 'uid:', self.get('uid'), 'cell:', to_str(self.get('cell')), '}')
        return result
# ============================================================
# ============================================================
    def Get_units_count(self):
        '''Возвращает кол-во юнитов в здании. Вызывается из C++'''
        return len(self._building_units)
# ============================================================
    def is_home(self):  # is home for units
        '''Является ли данное здание домом для юнитов, если да, то нужно переопределить это в классе-потомке'''
        return 0        # default, need to override for camp, etc...
# ============================================================
    def is_store(self):
        '''Является ли данное здание хранилищем ресурсов'''
        if self.status == 'normal': # только в нормальное здание можно нести ресурсы, а не в сломанное или только строящееся
            return self.GetAttrBool('handle_resources')
        else:
            return False
# ============================================================
    def need_base_resource(self):
        '''Требуется ли для данного здиния базовый ресурс, из которого идет производство основного'''
        return False
# ============================================================
    def get_status(self):
        return self.status
# ============================================================
    def Get_vacant_places_count(self):
        '''Кол-во свободных мест в домике, учитывая тех юнитов, которые еще только бегут. Вызывается из C++'''
        if not self.is_home():
            return 0
        _max = self.max_units_count
        _max = _max - len(self._building_units) - len(self.walking_home_units)# - len(self.home_sleep_units)
        return _max
# ============================================================
    def leave_unit(self, unit):
        '''Выпускает юнита из здания'''
        self._building_units.remove(unit)
        unit.show()
        trace('Building: leave_unit', unit.get('uid'), self.get('name'), self.get('uid'), unit.get('cell'))
# ============================================================
    def in_unit(self, unit):
        '''Заводит юнита в здание'''
        #self.remove_walking_home_unit(unit)
        if unit not in self._building_units:
            self._building_units.append(unit)
        unit.set_home(self)
        _is_home = self.is_home()
        if _is_home:
            _old_unit = unit
            unit = UMgr.add_free_unit_reface(unit)        # на всякий случай еще раз "освободим" юнита (если был занят) и перелицуем его, если это был не простой юнит
            unit.current_subtype = 'simple'
            if _old_unit != unit:
                unit.set_home(self)
                self._building_units.remove(_old_unit)
                self._building_units.append(unit)
                _old_unit = None
            TMgr.send_event('free_unit', unit)  # повторяем событие, чтобы сработали таски, которым нужен юнит simple-типа
        trace('Building: in_unit', unit.get('uid'), self.get('name'), self.get('uid'))
# ============================================================
    def add_walking_home_unit(self, unit):
        '''Добавляет юнита в список направляющихся домой. Эта инфа нужна для Get_vacant_places_count'''
        self.walking_home_units.append(unit)
        unit.home_building = self
# ============================================================
    def remove_walking_home_unit(self, unit):
        '''Исключает юнита из списка направляющихся домой'''
        self.walking_home_units.remove(unit)
# ============================================================
    def get_units_count(self):
        _count = 0
        for u in self._building_units:
            if u.is_in_home:
                _count += 1
        return _count
# ============================================================
    def get_first_unit(self):
        for u in self._building_units:
            if u.is_in_home:
                return u
        return None
# ============================================================
    def get_action_cell_name(self):
        '''Метод возвращает случайную клетку для юнита для строительства здания.'''
        _bcells = self.building_cells_names[:]  # работает ли???
        _has_cells = len(_bcells) > 0
        _name = self.get('name')
        for cell_name in self.reserved_building_cells_names:
            _bcells.remove(cell_name)
        for cell_name in self.building_cells_names: # убираем клетки, которые недоступны
            _cell = self.GetAttrCellFixed(cell_name)
            is_passable = self.IsCellPassable(_cell)
            if not is_passable:
                if cell_name in _bcells:
                    _bcells.remove(cell_name)
                    #trace('DBG: non passable cell', cell_name)
        length = len(_bcells)
        if length > 0:
            rnd_index = tools.Rand(0, length-1)
            return _bcells[rnd_index]
        elif _has_cells:    # если есть только непроходимые клетки (либо все зарезервированы)
            #trace('DBG had_cells:', self.get('name'), self.get('uid'))
            return None
        else:                   # если клетка не указана
            return ''
# ============================================================
    def get_attack_cell_name(self):
        '''Метод возвращает случайную клетку для дикаря для атаки здания.'''
        _bcells = self.barb_cells_names[:]
        for cell_name in self.barb_cells_names: # убираем клетки, которые недоступны
            _cell = self.GetAttrCellFixed(cell_name)
            if not self.IsCellAccessible(_cell):
                _bcells.remove(cell_name)
        length = len(_bcells)
        if length > 0:
            rnd_index = tools.Rand(0, length-1)
            return _bcells[rnd_index]
        else:
            return ''
# ============================================================
    def is_cell_passable(self, cell_name):
        '''Проходима ли ячейка здания'''
        _cell = self.GetAttrCellFixed(cell_name)
        _is_passable = self.IsCellAccessible(_cell)
        return _is_passable
# ============================================================
    def reserve_action_cell(self, cell_name):
        '''Резервирует строительную ячейку, чтобы она больше не выдвалась
        в качестве доступной для постройки'''
        if cell_name not in self.reserved_building_cells_names:
            self.reserved_building_cells_names.append(cell_name)
        pass
# ============================================================
    def unreserve_action_cells(self):
        '''Разрезервирует все зарезервированные ячейки'''
        self.reserved_building_cells_names = []
        #del self.reserved_building_cells_names[:]
# ============================================================
    def start_task_participation(self, task, sub = ''):
        '''Начало участия здания в задании'''
        _name = task.type+sub
        if _name in self.tasks_member:
            trace('ERROR: start_task_participation -> another task already participating', self.get('uid'))
            trace('   another:', _name, self.tasks_member[_name])
            trace('   current:', task)
        self.tasks_member[_name] = ('started', task)
        trace('start_task_participation:', self.get('name'), task.type, sub, self.get('uid'))
# ============================================================
    def stop_task_participation(self, task, sub = ''):
        '''Конец участия здания в задании'''
        _name = task.type+sub
        if _name not in self.tasks_member:
            trace( self.get('name') + " stop not participating " + _name, self.get('uid') )
            for t in self.tasks_member:
                trace('   but participating in:', t)
        else:
            self.tasks_member[_name] = None
            del self.tasks_member[_name]
        trace('stop_task_participation:', self.get('name'), task.type, sub, self.get('uid'))
# ============================================================
    def is_task_participating(self, type):
        '''Участвует ли уже здание в данном типе задания. Internal use only'''
        if type in self.tasks_member:
            if self.tasks_member[type] != None:
                return True
        return False
# ============================================================
    def get_participating_task(self, type):
        '''Если здание участвует в указанном задании, то возвращает указанное задание, иначе None.
            Если заданий одного типа будет несколько, то сохранится только последнее'''
        if type in self.tasks_member:
            if self.tasks_member[type] != None:
                return self.tasks_member[type][1]
        return None
# ============================================================
    def is_active(self, skip_priority_check = False):
        '''Возможны ли действия над объектом в данный момент. Метод может быть переопределен'''
        #return not self.HasAccess() and not self.is_deactivated and not self.non_selectable and self.get_next_level_name() # and self.is_next_level_possible()
        if not self.is_deactivated:
            if not self.non_selectable:
                if self.is_under_attack == 0:
                    #_state = self.SM.GetState()
                    #if _state not in ['awaiting_action','build','upgrade']:   # если подсветка только по кнопке
                    _is_builders_state = self.is_builders_state()
                    if not _is_builders_state:
                        _next_level_name = self.get_next_level_name(skip_priority_check) # and self.is_next_level_possible()
                        if _next_level_name:
                            return True
                    else:
                        return True
                    #elif self.is_mouse_over_zone == 'button':
                    #    return True
        return False
# ============================================================
    def is_active_raw(self):
        '''Не деактивировано и может селектиться'''
        #return not self.HasAccess() and not self.is_deactivated and not self.non_selectable and self.get_next_level_name() # and self.is_next_level_possible()
        return not self.is_deactivated and not self.non_selectable# and not self.lock_shown
# ============================================================
    def _cannot_repair_because_of_priority(self):
        '''Нельзя ремонтировать здание, потому что приоритеты мешают?'''
        if not self.is_deactivated:
            if not self.non_selectable:
                if self.status == 'broken':
                    if self.is_under_attack == 0:
                        _state = self.SM.GetState()
                        #if _state not in ['awaiting_action','build','upgrade']:   # если подсветка только по кнопке
                        _is_builders_state = self.is_builders_state()
                        if not _is_builders_state:
                            _template_name = self.get('name')
                            var_name = self.get('variant_name')
                            _can_be_repaired = BMgr.CanBeRepairedPriority(_template_name, self.get('locind'))
                            if not _can_be_repaired:
                                return True
                        else:
                            return False
        return False
# ============================================================
    _repairing_state_broken = ['upgrade', 'awaiting_action']
    _normal__place__broken = ['normal', 'place', 'broken']
    def is_functional(self):
        '''Здание, которое работает, в том смысле, что не сломано, не выжившее, нормально функционирующее'''
        _has_access = self.HasAccess()
        if not _has_access:
            return False
        _is_broken = self.is_broken()
        _b_state = self.SM.GetState()
        # ---
        if _is_broken:
            if _b_state not in self._repairing_state_broken:   # ремонтируемое здание считается функционирующим, поэтому их оставляем
                return False                                     # остальные скипуем
        if self.is_deactivated: return False
        #if self.is_under_attack: return False
        if self.status not in self._normal__place__broken:    # broken - мы выше уже отсекли просто поломанные, теперь только уже ремонтируемые, которые считаем рабочими
            return False # потому что place - это будущее рабочее здание
        return True
# ============================================================
    def is_functional_non_attacked(self):
        _functional = self.is_functional()
        if _functional:
            if self.is_under_attack:
                return False
            else:
                return True
        return False
# ============================================================
    def deactivate(self):
        _was_active = not self.is_deactivated
        self.is_deactivated = True
        self.is_selected = False
        if _was_active:
            self.on_non_functional()
        self.switch_animation()
# ============================================================
    def activate(self):
        _was_deactivated = self.is_deactivated
        self.is_deactivated = False
        if _was_deactivated:
            if self.status == 'normal':
                self.on_functional()
        self.switch_animation()
# ============================================================
    def get_subtask(self, parent_task, actor):
        '''Возвращает подзадание для юнита, который будет взаимодействовать с данным объектом'''
        return task_classes.Subtask(parent_task, actor)
# ============================================================
    def set_check(self, check):
        '''У зданий галки не ставятся'''
        pass
# ============================================================
    def on_worker_in(self, unit):
        '''Чтобы здание могло отловить когда пришел юнит-работник в здание'''
        self.PlaySound('SND_BUILDING_OCCUPY')
        self.update_next_res()
        self.switch_animation()
        _workers = len(self.worker_units)
        if _workers == 1:
            self._on_first_unit_in()
        self.FSM.HandleEvent('unit_in_event')
# ============================================================
# методы для отображения типса
# ============================================================
    _awaiting__build__upgrade = ['awaiting_action', 'build', 'upgrade']
    def is_green_state(self, skip_menu_check = False, _buttons = None):
        '''Цвет типса/селекта, с учетом состояния кнопок меню действий'''
        if not skip_menu_check:
            if not _buttons:
                _buttons = self.Get_menu_items_id()
            _buttons_count = len(_buttons)
            if _buttons_count > 0:
                _have_green = False
                for btn in _buttons:
                    _is_green = self.is_menu_item_green(btn)
                    if _is_green:
                        _have_green = True
                        break
                if not _have_green:
                    return False
        result = self._raw_is_green_state()
        return result
# ============================================================
    def get_green_menu_buttons_count(self, _buttons = None):
        '''Цвет типса/селекта, с учетом состояния кнопок меню действий'''
        _green_count = 0
        if not _buttons:
            _buttons = self.Get_menu_items_id()
        for btn in _buttons:
            _is_green = self.is_menu_item_green(btn)
            if _is_green:
                _green_count += 1
        return _green_count
# ============================================================
    def is_menu_button_green(self, btn_id, _buttons = None):
        '''Цвет типса/селекта, с учетом состояния кнопок меню действий'''
        if not _buttons:
            _buttons = self.Get_menu_items_id()
        for btn in _buttons:
            if btn_id == btn:
                _is_green = self.is_menu_item_green(btn)
                if _is_green:
                    return True
        return False
# ============================================================
    def _raw_is_green_state(self):
        '''Цвет типса/селекта, не учитывает условия меню действий'''
        if self.cached_raw_green_state == None:
            self.cached_raw_green_state = self._raw_is_green_state_int()
        return self.cached_raw_green_state
# ============================================================
    def _raw_is_green_state_int(self):
        '''Цвет типса/селекта, не учитывает условия меню действий'''
        if self.non_selectable: # если не выделяется, то нет смысл в красном типсе
            return True
        state = self.SM.GetCurrentState()
        if self.is_builders_state():
            return True # здесь нужен зеленый типс
        _has_access = self.HasAccess()
        if self.is_deactivated:
            if not _has_access:
                return False
            else:
                return True
        if not _has_access:
            return False
        _has_barbarians = UStats.GetLevelValue('barbarian_attack') > 0
        if self.status in self._broken__place__map_place: _has_barbarians = False
        if self.subtype == 'watchtower': _has_barbarians = False
        if _has_barbarians:
            if self.status == 'normal':
                if self.has_next_level: # если апгрейдабельное здание
                    return False
        _is_active = self.is_active()
        if self.has_next_level:
            _name = self.get('name')
            _need_worker = not self.GetAttrBool('need_no_worker')
            if not self.clock_shown:
                map_index = self.get('locind')
                _build_possible = BMgr.Is_build_action_possible(map_index, _name, _need_worker)
                if not _build_possible:
                    return False
        if _is_active:
            return True
        if self.status == 'map_place':
            return True
        state = self.SM.GetCurrentState()
        if self.lock_shown:
            return True
        self.update_res_enough_state(False)  # обновим состояние ресурсов
        # нужно:
        # - проверить доступность апгрейда по кол-ву юнитов (п.97)
        # - автообновление статуса для цвета стрелок...
        return True
# ============================================================
    def Get_tip_image(self):
        '''Возвращает ID изображения, которое используется для бэка типса. Может переопределяться у зданий с полями'''
        # чтобы обновить кадр селекта:
        _frame = 0
        _is_green_state = self.is_green_state()
        if not _is_green_state: _frame = 1
        drawer = self.Drawer
        drawer.SetFrame('select', _frame)
        drawer.SetSequence('arrow', _frame)    # номер последовательности сейчас совпадает с номером кадра
        # ---
        if not _is_green_state:   # если хватает ресурсов на следующий апгрейд
            return 'IMG_I_TIP_BG_RED'
        return 'IMG_I_TIP_BG'
# ============================================================
    def show_select(self, show):
        self.Drawer.ShowImages('select', show)
# ============================================================
    def Get_tip_text(self, by_click = False):
        '''Возвращает LIST, содержащий текстовые ID'''
        result = []
        if not self._get_tips_extra('before_name', result, by_click): return result
        item_name = self.GetAttrStr('item_name')
        if not item_name:
            return result
        if item_name == 'NO_TIPS':
            return result
        _has_access = self.HasAccess()
        _fsm_state = self.FSM.GetStateID()
        state = self.SM.GetCurrentState()
        if not by_click:
            _is_cancel_button_mode = self.is_cancel_button_mode(_fsm_state) # sloooow method
            if _is_cancel_button_mode:
                if self.is_mouse_over_zone == 'button':
                    result.append( ('text', 'COLOR_HEADER') )
                    result.append( ('text', 'CLICK_TO_CANCEL') )
                    return result
            result.append( ('text', 'COLOR_HEADER') )
            result.append( ('text', item_name) )
            if self.is_deactivated:
                if state in self._normal__empty:
                    return result
            # перенес код для отображения того, что дает здание:
            self._append_text_willgive_with_conditions(state, result)
            if not self._get_tips_extra('after_name', result, by_click): return result
            if state in self._normal__empty:
                if self.status == 'normal':
                    self._get_gives(result)
            if not self._get_tips_extra('after_gives', result, by_click): return result
            self._append_effectivness(result)
            self._append_genbar(result)
            self._append_lifebar(result)
            if not self.is_active():
                self._append_bonus_info(result)
                return result
            if state in self._normal__empty:
                _has_barbarians = UStats.GetLevelValue('barbarian_attack') > 0
                if self.status in self._broken__place__map_place:
                    _has_barbarians = False # на них варвары не действуют
                if self.subtype == 'watchtower': _has_barbarians = False
                if self.has_next_level and self.status not in self._map_place__place:
                    if not self._get_tips_extra('before_upgrade_cost', result, by_click): return result
                    res = self.next_res
                    if res:
                        if not self._get_tips_extra('right_before_upgrade_cost', result, by_click): return result
                    _name = self.get('name')
                    _need_worker = not self.GetAttrBool('need_no_worker')
                    map_index = self.get('locind')
                    if not BMgr.Is_build_action_possible(map_index, _name, _need_worker): # если недостаточно юнитов
                        result.append( 'newline' )
                        result.append( ('text', 'COLOR_RED') )
                        result.append( ('text', 'NOT_ENOUGH_UNITS') )
            else:
                self._append_pbar(result)   # добавим прогрессбар, если нужно
            self._append_bonus_info(result)
            if not self._get_tips_extra('end', result, by_click): return result
            _is_debug = tools.IsDebug()
            if _is_debug: # добавить данные по навешанным на здания таскам
                if self.tasks_member:
                    result.append( 'newline' )
                    for tname in self.tasks_member:
                        result.append( ('debug', tname) )
                        result.append( ('debug', ', ') )
        else:
            if not _has_access:
                result = []
                result.append( ('text', 'COLOR_LINES') )
                result.append( ('text', 'NO_PATH') )
                return result
            if self.is_deactivated:
                _not_accessible = False
                _is_yard = self.IsSubTypeOf(['verf', 'boatyard'])
                if not _is_yard: _not_accessible = True
                if self.status == 'broken': _not_accessible = True
                if _not_accessible:
                    result.append( ('text', 'COLOR_LINES') )
                    #result.append( ('text', ('NOT_NEED_YET','NOT_ACCESSIBLE_YET')[self.status == 'broken']) )
                    result.append( ('text', 'NOT_ACCESSIBLE_YET') )
                    return result
                else:
                    return self.Get_tip_text(False)
            _is_fixed_place = self.GetAttrBool('fixed_place')
            if _is_fixed_place:
                if self.status == 'map_place':
                    result.append( ('text', 'COLOR_LINES') )
                    result.append( ('text', 'MAP_PLACE') )
                    result.append('newline')
                    result.append( ('text', 'COLOR_HEADER') )
                    _item_map_name = self.GetAttrStr('item_map_name')
                    result.append( ('text', _item_map_name) )
                    return result
            _is_active = self.is_active()
            if not _is_active:
                if self.non_selectable:
                    return self.Get_tip_text(False) # если по клику нет типса, то вернем его как без клика
                building_cannot_repair_because_of_priority = self._cannot_repair_because_of_priority()
                if building_cannot_repair_because_of_priority:
                    result.append( ('text', 'COLOR_LINES') )
                    #result.append( ('text', ('NOT_NEED_YET','NOT_ACCESSIBLE_YET')[self.status == 'broken']) )
                    result.append( ('text', 'NOT_ACCESSIBLE_YET') )
                    return result
                else:
                    return self.Get_tip_text(False) # если по клику нет типса, то вернем его как без клика
            _is_cancel_button_mode = self.is_cancel_button_mode(_fsm_state)
            if _is_cancel_button_mode:
                if self.is_mouse_over_zone == 'button':
                    result.append( ('text', 'COLOR_HEADER') )
                    result.append( ('text', 'CLICK_TO_CANCEL') )
                    return result
            if state in self._normal__empty:
                if self.has_next_level:
                    _name = self.get('name')
                    _need_worker = not self.GetAttrBool('need_no_worker')
                    map_index = self.get('locind')
                    _is_build_action_possible = BMgr.Is_build_action_possible(map_index, _name, _need_worker) # если недостаточно юнитов
                    if not _is_build_action_possible:
                        result.append( ('text', 'COLOR_RED') )
                        result.append( ('text', 'NOT_ENOUGH_UNITS') )
                        return result
                    _has_barbarians = UStats.GetLevelValue('barbarian_attack') > 0
                    if self.subtype == 'watchtower': _has_barbarians = False
            if not self._get_tips_extra('end_c', result, by_click): return result
            if not result:
                return self.Get_tip_text(False) # если по клику нет типса, то вернем его как без клика
        return result
# ============================================================
    def _append_repair_hack_text(self, result, hack_text):
        '''Добавляет специальные строчки к типсу'''
        result.append('newline')
        result.append( ('text', 'COLOR_LINES') )
        result.append( ('text', hack_text) )
# ============================================================
    def _append_text_willgive_with_conditions(self, state, result):
        if state in self._normal__empty:
            _has_barbarians = UStats.GetLevelValue('barbarian_attack') > 0
            if self.status in self._broken__place__map_place:
                _has_barbarians = False # на них варвары не действуют
            if self.subtype == 'watchtower': _has_barbarians = False
            if not _has_barbarians:
                if self.has_next_level:
                    if self.status not in self._map_place__place:
                        res = self.next_res
                        if res:
                            self._append_text_willgive(result)
# ============================================================
    willgive_buildings = {'granitemine':0, 'silvermine':0, 'fisherman':0, 'farm':0,\
                                'woodcutter':0, 'quarry':0, 'forester':0}
    def _append_text_willgive(self, result):
        '''Добавляет к описанию что дает апгрейд/ремонт'''
        _stype = self.get('stype')
        _repair_hack = False
        _repair_hack_text = ''
        if self.status == 'broken':
            _repair_hack = self.GetAttrBool('repair_tip_hack')
            _repair_hack_text = self.GetAttrStr('repair_tip_text')
        # ##
        if _stype in self.willgive_buildings:
            next_var_name = self.get_next_level_name()
            if next_var_name in ['level_2', 'level_3']:
                if not _repair_hack:
                    result.append('newline')
                    result.append( ('text', 'COLOR_LINES') )
                    result.append( ('text', ('UPGRADE_GIVES', 'REPAIR_GIVES')[self.status == 'broken']) )
                    result.append('newline')
                    result.append( ('text', ('DOUBLE_PERFORMANCE', 'TRIPLE_PERFORMANCE')[next_var_name == 'level_3']) )
                else:
                    self._append_repair_hack_text(result, _repair_hack_text)
            elif _repair_hack:
                self._append_repair_hack_text(result, _repair_hack_text)
        elif _stype in ['totem', 'totem_stone']:
            next_var_name = self.get_next_level_name()
            if next_var_name:
                if not _repair_hack:
                    result.append('newline')
                    result.append( ('text', 'COLOR_LINES') )
                    result.append( ('text', ('UPGRADE_GIVES', 'REPAIR_GIVES')[self.status == 'broken']) )
                    result.append('newline')
                    result.append( ('text', 'INCREASE_TERRITORY') )
                else:
                    self._append_repair_hack_text(result, _repair_hack_text)
        elif _stype == 'watchtower':
            next_var_name = self.get_next_level_name()
            if next_var_name:
                if not _repair_hack:
                    result.append('newline')
                    result.append( ('text', 'COLOR_LINES') )
                    result.append( ('text', ('UPGRADE_GIVES', 'REPAIR_GIVES')[self.status == 'broken']) )
                    result.append('newline')
                    result.append( ('text', 'INCREASE_TERRITORY') )
                    if next_var_name == 'level_2':
                        result.append('newline')
                        result.append( ('text', 'PROTECTION_AGAINST_BARBARIAN') )
                    elif next_var_name == 'level_3':
                        result.append('newline')
                        result.append( ('text', 'FULL_PROTECTION_AGAINST_BARBARIAN') )
                else:
                    self._append_repair_hack_text(result, _repair_hack_text)
        elif _stype == 'warehouse':
            next_var_name = self.get_next_level_name()
            if next_var_name:
                if next_var_name in ['level_2', 'level_3']:
                    if not _repair_hack:
                        result.append('newline')
                        result.append( ('text', 'COLOR_LINES') )
                        result.append( ('text', ('UPGRADE_GIVES', 'REPAIR_GIVES')[self.status == 'broken']) )
                        result.append('newline')
                        if next_var_name == 'level_2':
                            result.append( ('text', 'PROTECTION_BARB_1') )
                        elif next_var_name == 'level_3':
                            result.append( ('text', 'PROTECTION_BARB_2') )
                    else:
                        self._append_repair_hack_text(result, _repair_hack_text)
                elif _repair_hack:
                    self._append_repair_hack_text(result, _repair_hack_text)
        elif _stype == 'house':
            next_var_name = self.get_next_level_name()
            if next_var_name:
                if self.status == 'normal':
                    result.append('newline')
                    result.append( ('text', 'COLOR_LINES') )
                    result.append( ('text', 'UPGRADE_GIVES') )
                    result.append('newline')
                    result.append( ('text', 'ADDITIONAL_UNITS') )
                elif self.status == 'broken':
                    if not _repair_hack:
                        result.append('newline')
                        result.append( ('text', 'COLOR_LINES') )
                        result.append( ('text', 'REPAIR_GIVES') )
                        result.append('newline')
                        result.append( ('text', 'ADDITIONAL_UNITS_2') )
                    else:
                        self._append_repair_hack_text(result, _repair_hack_text)
        elif _stype in ['hq', 'hq_small']:
            next_var_name = self.get_next_level_name()
            if next_var_name:
                if self.status == 'normal':
                    result.append('newline')
                    result.append( ('text', 'COLOR_LINES') )
                    result.append( ('text', 'UPGRADE_GIVES') )
                    result.append('newline')
                    result.append( ('text', 'ADDITIONAL_UNITS') )
        elif _repair_hack:  # если галка установлена и статус сломанного здания
            self._append_repair_hack_text(result, _repair_hack_text)
        elif self.status == 'broken':
            _res = self._get_willgive_res()
            if _res:
                result.append('newline')
                result.append( ('text', 'COLOR_LINES') )
                result.append( ('text', 'WILL_GIVE') )
                result.append('space')
                _res_numbers = self._append_res_values(_res)
                result.extend( _res_numbers )
# ============================================================
    def _get_tips_extra(self, place, result, by_click):
        '''Предназначается для переопределения в потомках
        place: метка позиции вставки типса
        result: список значений, в который нужно добавлять данные
        by_click: типс по клику или нет
        returns: False - для завершения типса, т.е. после него ни одной строки не прибавится'''
        return True
# ============================================================
    def _get_gives(self, result):
        '''Возвращает текст для части типса GIVES'''
        _gives = self.GetRes('normal_gives_1')
        if _gives:
            if not _gives.Empty():
                _gives = _gives.Clone()
                #_gives.FixFood()        # чтобы еда учитывалась верно
                _need_double = self.Signals.Has('bonus_gen_double')
                if not _need_double: _need_double = self.is_generated_res_doubled
                if _need_double: _gives.MultiplyValues(2)
                #if not _gives.MainResEmpty():
                result.append( 'newline' )
                result.append( ('text', 'COLOR_LINES') )
                result.append( ('text', 'GIVES') )
                result.append( 'space' )
                res_numbers = self._append_res_values(_gives)
                result.extend( res_numbers )
# ============================================================
    def _append_bar(self, result, _progress, bar_images, _places_count = 10):
        '''Добавляет bar r result по указанным параметрам'''
        _filled_places = _places_count * _progress
        _filled_places = int(_filled_places + 0.5)
        if _filled_places > _places_count: _filled_places = _places_count
        elif _filled_places < 0: _filled_places = 0
        result.append( ('text', bar_images.left) ) #'TIP_BAR_L'
        _bar_full = bar_images.full
        _bar_empty = bar_images.empty
        for i in range(_filled_places):
            result.append( ('text', _bar_full) ) #'TIP_BAR_FULL'
        for i in range(_places_count-_filled_places):
            result.append( ('text', _bar_empty) ) #'TIP_BAR_EMPTY'
        result.append( ('text', bar_images.right) ) #'TIP_BAR_R'
# ============================================================
    _build__upgrade = ['build', 'upgrade']
    _bar_progress_images_dict = {'left':'TIP_BAR_L', 'right':'TIP_BAR_R', 'full':'TIP_BAR_FULL', 'empty':'TIP_BAR_EMPTY'}
    def _append_pbar(self, result):
        state = self.SM.GetState()
        if state in self._build__upgrade:
            result.append('newline')
            if state == 'build':
                result.append( ('text', 'COLOR_LINES') )
                result.append( ('text', 'TIPS_BUILDING') )
                result.append('newline')
            elif state == 'upgrade':
                if not self.IsSubTypeOf(['verf', 'boatyard']) or self.status == 'broken':
                    result.append( ('text', 'COLOR_LINES') )
                    result.append( ('text', ('TIPS_UPGRADING', 'TIPS_REPAIRING')[self.status == 'broken'] ) )
                    result.append('newline')
            _progress = self.build_progress
            if self.build_time > 0: _progress /= self.build_time
            else: _progress = 0
            self._append_bar(result, _progress, self._bar_progress_images_dict)
# ============================================================
    _bar_eff_images_dict = {'left':'TIP_EFFBAR_L', 'right':'TIP_EFFBAR_R', 'full':'TIP_EFFBAR_FULL', 'empty':'TIP_EFFBAR_EMPTY'}
    _fields = ['farm_field', 'forester_field']
    def _append_effectivness(self, result):
        stype = self.get('stype')
        if self.is_deactivated:
            return
        if self.is_indians:
            return
        _need_no_worker = self.GetAttrBool('need_no_worker')
        if _need_no_worker:
            if stype not in self._fields:
                return
        if self.status in self._broken__place__map_place:
            return
        state = self.SM.GetState()
        if state in ['build', 'upgrade', 'repair', 'awaiting']:
            return
        if stype in self.has_no_effectiveness:
            return
        if stype in ['farm', 'forester']:
            return
        state = self.SM.GetState()
        if state in self._normal__empty:
            if self.status == 'normal':
                _coeff = self.get_effectiveness_coeff()
                result.append('newline')
                result.append( ('text', 'COLOR_LINES') )
                result.append( ('text', 'TIP_EFFECTIVNESS') )
                result.append('newline')
                self._append_bar(result, _coeff, self._bar_eff_images_dict)
        #if tools.IsDebug():
        #    result.append( ('digit', self.get_effective_gen_time()) )
# ============================================================
    _bar_life_images_dict = {'left':'TIP_LIFEBAR_L', 'right':'TIP_LIFEBAR_R', 'full':'TIP_LIFEBAR_FULL', 'empty':'TIP_LIFEBAR_EMPTY'}
    def _append_lifebar(self, result):
        if self.energy >= 100 or self.energy == 0:
            return
        if self.is_indians:
            return
        _need_no_worker = self.GetAttrBool('need_no_worker')
        if _need_no_worker:   # эти здания вроде как все нельзя атаковать (но ниже они перечислены)
            return
        stype = self.get('stype')
        if stype in self.not_for_attack:
            return
        if self.status in self._broken__place__map_place:
            return
        if self.is_under_attack:
            result.append('newline')
            result.append( ('text', 'COLOR_LINES') )
            result.append( ('text', 'TIPS_BREAKING') )
        result.append('newline')
        result.append( ("text", "COLOR_LINES") )
        result.append( ('text', 'TIP_LIFE') )
        result.append('newline')
        self._append_bar(result, self.energy/100, self._bar_life_images_dict)
# ============================================================
    _bar_gen_images_dict = {'left':'TIP_GENBAR_L', 'right':'TIP_GENBAR_R', 'full':'TIP_GENBAR_FULL', 'empty':'TIP_GENBAR_EMPTY'}
    def _append_genbar(self, result):
        '''Бар производства на типсе'''
        if self.is_indians:
            return
        if not self.is_generating_res:
            return
        _progress = self.Signals.GetProgress('resource_generated')
        if _progress < 0:
            return
        result.append('newline')
        result.append( ('text', 'COLOR_LINES') )
        result.append( ('text', 'TIP_RES_GENERATION') )
        result.append('newline')
        self._append_bar(result, _progress, self._bar_gen_images_dict)
# ============================================================
    def _append_res_values(self, resources, need_comma = True):
        '''Для типса. Генерирует список со значениями переданного ресурса, с соответствующими изображениями'''
        result = []
        if not resources: return result
        first_res = True
        for res_name in self.resource_names:    # выводим значения ресурсов...
            res_val = resources.Get(res_name)
            if res_val > 0:     # на всякий случай
                if not first_res:
                    if need_comma:
                        result.append( 'comma' )
                    result.append( 'space' )
                first_res = False
                result.append( ('digit', resources.Get(res_name)) )
                result.append( ('image', self.res_images[res_name]) )
        return result
# ============================================================
    _non_bonus_states_dict = {'prenormal':0, 'normal':0, 'upgrading':0,
                                'generating':0, 'generated':0, 'preremoving':0}
    def _append_bonus_info(self, result):
        '''Для типса. Добавляет подпись о действующих бонусах'''
        _fsm_state = self.FSM.GetStateID()
        if _fsm_state not in self._non_bonus_states_dict:
            return result
        if self.Signals.Has('bonus_gen_double'):
            result.append('newline')
            result.append( ('text', 'COLOR_LINES') )
            result.append( ('text', 'UNDER_BONUS_DOUBLE_RES') )
        if self.Signals.Has('bonus_restore_eff'):
            result.append('newline')
            result.append( ('text', 'COLOR_LINES') )
            result.append( ('text', 'UNDER_BONUS_EFFECTIVENESS') )
        return result
# ============================================================
    def Get_tip_position(self, tip_size, camera_pos, window_rect, zoom):
        '''Возвращает координаты типса от левого-верхнего угла картинки типса'''
        return toolz.Tools.get_tip_position(self, tip_size, camera_pos, window_rect)
# ============================================================
    def has_workers(self):
        return len(self.worker_units) > 0
# ============================================================
    def get_worker(self):
        if self.has_workers():
            return self.worker_units[0]
        else:
            return None
# ============================================================
    def can_build(self):
        '''Можно ли строить здание - доступны ли юниты'''
        _name = self.get('name')
        _need_worker = not self.GetAttrBool('need_no_worker')
        map_index = self.get('locind')
        can_build = BMgr.Is_build_action_possible(map_index, _name, _need_worker)
        return can_build
# ============================================================
    def is_menu_item_green(self, item_id):
        if item_id not in self.cached_menu_item_green:
            self.cached_menu_item_green[item_id] = self._is_menu_item_green_int(item_id)
        return self.cached_menu_item_green[item_id]
# ============================================================
    def _is_menu_item_green_int(self, item_id):
        _need = False
        _extra = ''

        if item_id == 'upgrade':
            _need, _extra = self._need_upgrade_button(True)
        elif item_id == 'repair':
            _need, _extra = self._need_repair_button(True)
        elif item_id == 'restore_eff':
            _need, _extra = self._need_restore_eff_button(True)
        elif item_id == 'remove':
            _need, _extra = self._need_remove_button(True)
        if _need:
            if _extra: return False
            else: return True
        return False
# ============================================================
    def get_raw_button_frame(self): # sloooow method...
        '''Возвращает номер кадра кнопки, без учета положения курсора мыши'''
        frame = -1
        if self.is_deactivated:
            return -1
        _fsm_state = self.FSM.GetStateID()
        if _fsm_state in self.no_button_states:
            return -1
        _has_barbarians = UStats.GetLevelValue('barbarian_attack') > 0
        if _has_barbarians:
            return -1
        if _fsm_state == 'preremoving':
            return 3
        _is_wait_worker = self.is_wait_worker_state(_fsm_state)
        if not _is_wait_worker:     # slow part begins...
            _buttons = self.Get_menu_items_id()
            _buttons_count = len(_buttons)
            if _buttons_count == 0:
                return -1
            elif _buttons_count > 0:
                _is_green_state = self.is_green_state(False, _buttons)
                if not _is_green_state:
                    return 11
                _green_count = self.get_green_menu_buttons_count(_buttons)
                if _green_count == 1: # если всего один пункт меню
                    _is_remove_green = self.is_menu_button_green('remove', _buttons)
                    if _is_remove_green: # и если это пункт сноса, то
                        return 18       # желтая кнопка
                return 9
        if _is_wait_worker:    # ожидание юнита для строительства/апгрейда/ремонта
            frame = 3
        #elif self.status == 'place':
        #    frame = 3
        return frame
# ============================================================
    def is_cancel_button_mode(self, _fsm_state):
        if self.is_mouse_over:
            if _fsm_state not in self._non_cancelable_sates:   # если подсветка только по кнопке
                frame = self.get_raw_button_frame()
                if frame in self.button_active_frames: # подсветка кнопки при наведении
                    if frame == self.cancel_build_active_frame:     # если кнопка отмены
                        return True
        return False
# ============================================================
    def Get_button_frame(self):
        _fsm_state = self.FSM.GetStateID()
        frame = self.get_raw_button_frame()
        if frame >= 0:
            if self.is_mouse_over:
                if _fsm_state not in self._non_cancelable_sates:
                    if frame in self.button_active_frames: # подсветка кнопки при наведении
                        if frame == self.cancel_build_active_frame:     # для отмены подсвечиваем только над кнопкой
                            if self.is_mouse_over_zone == 'button':
                                frame += 1
                        else:
                            frame += 1
        if frame < 0:
            self.Drawer.ShowImages('button', False)
        else:
            self.Drawer.ShowImages('button')
        return frame
# ============================================================
    def send_extra_event(self):
        if self.extra_event:
            BMgr.SendGlobalMsg('extra_event', self.extra_event)
            self.extra_event = ''
# ============================================================
    def get_needs_res(self, variant_name):  #@opt вынести в C++ полностью?
        need_res_name = self.GetAttrStr('need_res_name')
        if need_res_name == '':
            need_res_name = 'normal_needs_1'
        return self.GetRes(variant_name, need_res_name)
# ============================================================
    def get_res(self, attr_resname, default_resname):
        _resname = self.GetAttrStr(attr_resname)
        if _resname == '':
            _resname = default_resname
        return self.GetRes(_resname)
# ============================================================
    def _get_willgive_res(self):
        next_var_name = self.get_next_level_name()
        if next_var_name:
            _res = self.GetRes(next_var_name, 'normal_gives_1')
            _is_res_empty = True
            if _res:
                _is_res_empty = _res.Empty()
            if not _is_res_empty:
                return _res
        return None
# ============================================================
    def Force_highlight(self, highlight):
        pass
# ============================================================
    attack_blink_layers = ['default', 'floor', 'bush1', 'bush2', 'bush3', 'bush4', 'bush5', 'bush6']
    def action_hit_me(self, count=1):
        '''По зданию сделан удар
        count: сколько ударов за раз'''
        if self.status == 'broken':
            return
        if self.energy <= 0:     # на всякий случай, чтобы объект нельзя было разрушить дважды
            return
        #self.action_was_started = True  # чтобы уже нельзя было отменить
        self.energy -= count * self.energy_hit # по 10% за удар
        #self.action_progress = 1 - self.action_hits_last / self.action_hits_count

        self.Drawer.Blink(self.attack_blink_layers, 1, 0xffff8888, 0.5)
        self.on_action_hit_me()        # чтобы можно было переопределить
        self.scale_bump()
        if self.energy <= 0:
            #self.action_was_started = False # чтобы объект работал, если у него сложная логика
            #effect_digged_name = self.GetAttrStr('effect_complete_name')
            #if effect_digged_name:
            #    self.AddEffect(effect_digged_name, self.get('position'))
            self.on_action_hits_me_complete()
# ============================================================
    def on_action_hit_me(self):
        '''Для переопределения'''
        pass
# ============================================================
    def on_action_hits_me_complete(self):
        # нужно поломать здание
        _vname = self.get('variant_name')
        if _vname in self.broken_switch:
            _next_name = self.broken_switch[_vname]
            self.show_arrow(False)
            trace('on_action_hits_me_complete:')
            trace(' on_action_hits_me_complete:', self.get('name'), self.get('uid'), _next_name)
            self.SwitchLevel(_next_name)   # переключаемся на следующий вариант
            self.AddEffect('brocken_smoke', self.get('position'))   # после разрушения варварами
            self.PlaySound('SND_BUILDING_KILLED')
            self.status = self.GetAttrStr('status')
            self.Signals.Cancel('resource_generated')       # ресурсы больше не генерим
            self.update_next_res()
            self.unreserve_action_cells()
            BMgr.SendGlobalMsg('buildings_count_change')
            #letter = []
            #letter.append( ('text', 'MSG_BUILDED') )
            #letter.append( 'space' )
            #item_name = self.GetAttrStr("item_name")
            #letter.append( ('text', item_name) )
            #EMgr.AppendLetter(letter, self.get('uid'))
            self.update_res_enough_state()
            self.is_under_attack = 0
            #TMgr.cancel_task_by_uid(self.task_call_res_uid)
            #self.task_call_res_uid = 0
            self.stop_resource_generation()     # остановим генерацию, если сломано здание
            self.switch_animation()
            self._on_broken()
            self.on_switch_level_complete()
            UI.HideTopText()
            UI.show_top_text2('BARB_BUILDING_BROKEN', self.get('locind'), self.get('uid'))
# ============================================================
    def on_attack_started(self):
        UI.workers_texts = [] # обрубаем все сообщения
        UI.HideTopText()
        UI.show_top_text2('BARBARIAN_ATTACK_BUILDING', self.get('locind'), self.get('uid'), True)
# ============================================================
    def on_attack_canceled(self):
        '''Вызывается при отмене атаки одним из дикарей'''
        #if self.is_under_attack == 0:   # если больше атакующих нет, то
        #    self.start_resource_generation()    # запускаем генерацию ресурса, либо дровосека/камнелома
        pass
# ============================================================
    def return_worker_home(self):
        has_workers = len(self.worker_units)
        if has_workers:
            unit = self.worker_units.pop(0)
            UMgr.remove_building_worker(unit)  # уменьшим число занятых зданиями работников
            unit.start_work()       # освобождаем юнита от домика
            #unit.set_current_subtype('quarryman')
            unit.set_current_subtype('simple')
            unit.show()       # освобождаем юнита от домика
            UMgr.add_free_unit(unit) #если заняли юнита делом, то сообщим об этом
            #self.is_unit_inside = False
            self.switch_animation()
            return unit
            #self.PlaySound('SND_BUILDING_ABANDONED')
        return None
# ============================================================
# ============================================================
    def get_action_text(self, for_task_slide = False):
        return ''
# ============================================================
    def start_comet(self, _name, result_signal = 'comet_fly_complete'):
        '''Запускает комету, если указана в атрибутах, иначе просто дает сигнал на завершение удаления'''
        _comet_effect_name = _name
        if not _comet_effect_name:
            _comet_effect_name = self.GetAttrStr('comet_name')
        if _comet_effect_name:
            LBoard.StartComet(self, _comet_effect_name, result_signal)
            self.PlaySoundG('SND_ITEM_FLIGHT')
        else:
            self.Signals.Restart(result_signal, 1)
# ============================================================
    def on_before_action_started(self): # for safe only
        pass
# ============================================================
    def Get_workers_count(self):
        return len(self.worker_units)
# ============================================================
    def get_boost_multiplier(self):
        return 1.0 + self.GetBoostMultiplier() * (self.boost_work_percent/100.0)
# ============================================================
    def get_loot(self):
        '''Заглушка безопасности, на случай, если вызовут этот метод у здания (хотя не должно быть такого)'''
        return []
# ============================================================
    def start_final_countdown(self):
        # start round bar
        self.Signals.Restart('remove_countdown', 3000)
        self.StartRoundBar(3000)
        self.PlaySound('SND_DESTROY_COUNTDOWN')
        _uid = self.get('uid')
        if _uid == '4449':
            EMgr.SendGlobalMsg('preremove_started_4449')
# ============================================================
    def cancel_countdown(self):
        self.Signals.Cancel('remove_countdown')
        self.StopRoundBar()
# ============================================================
    def add_afterdeath_item(self):
        '''Добавляет воронку под здание'''
        _cell = self.GetAttrCellFixed('afterdeath_cell')
        _crater_template_name = self.GetAttrStr('crater_template', 'crator1')
        # добавить воронку под здание
        _new_item = BMgr.AddItemToMap(self.get('locind'), _crater_template_name, '', _cell)
        #_new_item.Drawer.ShowOut('', 10000)
        BMgr.remove_item_uid(_new_item)
# ============================================================
    def drop_cost_items(self, free_unit = None):
        '''Выбрасывает на поле объекты с остаточной стоимостью здания'''
        _res = self.get_res('remove_resname', 'remove_res')
        _food = _res.Get('food')
        _boards = _res.Get('boards')
        _bricks = _res.Get('bricks')
        _coins = _res.Get('coins')
        _params = [ ('food1', 'rem_food_cell', 'food', _food),
                    ('woods', 'rem_boards_cell', 'boards', _boards),
                    ('bricks', 'rem_bricks_cell', 'bricks', _bricks),
                    ('coins', 'rem_coins_cell', 'coins', _coins)]
        for _p in _params:
            _res_item = self.drop_item(_p[0], '', _p[1], _p[2], _p[3], free_unit, True)
            if _res_item: _res_item.set_shaded(True)
# ============================================================
    def drop_item(self, _template, _variant, _cell_name, _res_name, _count, _free_unit, _create_task = False):
        _new_item = None
        if _count > 0:
            _cell = self.GetAttrCellFixed(_cell_name)
            _new_item = BMgr.AddItemToMap(self.get('locind'), _template, _variant, _cell)
            if not _new_item:
                return
            _new_item.completeGivesRes = _new_item.completeGivesRes.Clone()
            _new_item.completeGivesRes.Set(_res_name, _count)
            _pos = _new_item.GetAttrPoint('center_item_pos')
            _new_item.Drawer.ResIn(_new_item.item_layers, _pos, 1000)
            if _create_task:
                if _free_unit:
                    if not _free_unit.task:
                        _item_actor = _new_item.GetAttrStr('item_actor')
                        if _item_actor != 'general':
                            _new_item.task = task_walkw.WalkWorkerTask(_new_item, _free_unit)
                            TMgr.add_task( _new_item.task )
                            return _new_item
                _new_item.on_click('default')
            return _new_item
# ============================================================
    def Get_menu_item_header_text(self):
        return [('text', self.GetAttrStr('item_name'))]
# ============================================================
    def is_clickable(self, skip_priority_check = False):
        '''Объект вообще можно кликнуть?'''
        _partial_fog = self.Drawer.IsPartiallyInFog()
        if _partial_fog:        # если в тумане
            return (False, '')
        if self.is_deactivated:
            return (False, '')
        _has_access = self.HasAccess()
        if not _has_access:
            return (False, '')
        # --------
        if skip_priority_check:
            if self.cached_is_clickable_skip == None:
                self.cached_is_clickable_skip = self._is_clickable_int(skip_priority_check)
            return self.cached_is_clickable_skip
        else:
            if self.cached_is_clickable_noskip == None:
                self.cached_is_clickable_noskip = self._is_clickable_int(skip_priority_check)
            return self.cached_is_clickable_noskip
# ============================================================
    def _is_clickable_int(self, skip_priority_check = False):
        '''Объект вообще можно кликнуть?'''
        _partial_fog = self.Drawer.IsPartiallyInFog()
        if _partial_fog:        # если в тумане
            return (False, '')
        if self.is_deactivated:
            return (False, '')
        _has_access = self.HasAccess()
        if not _has_access:
            return (False, '')
        _is_active = self.is_active(skip_priority_check)
        _fsm_state = self.FSM.GetStateID()
        _has_remove_or_restore = self.has_remove_or_restore(_fsm_state)
        if not _is_active:
            if _has_remove_or_restore:
                _is_active = True
        if not _is_active:
            return (False, '')
        if self.is_under_attack > 0:    # если здание атакуют
            return (False, 'barbarian_attack')
        _has_barbarians = UStats.GetLevelValue('barbarian_attack') > 0
        if self.status in self._broken__place__map_place: _has_barbarians = False
        if self.subtype == 'watchtower': _has_barbarians = False
        if _has_barbarians:
            return (False, 'barbarian_attack')
        _is_no_button_state = self.is_no_button_state()
        if _is_no_button_state:
            return (False, '')
        return (True, '')
# ============================================================
    def is_no_button_state(self):
        #state = self.SM.GetCurrentState()
        _fsm_state = self.FSM.GetStateID()
        #if _fsm_state == 'wait_unit':
        #    return False
        if _fsm_state in self.no_button_states:
            return True
        return False
# ============================================================
    def is_not_enough_workers_for_actions(self):
        _name = self.get('name')
        _need_worker = not self.GetAttrBool('need_no_worker')
        map_index = self.get('locind')
        if not BMgr.Is_build_action_possible(map_index, _name, _need_worker):
            return True
        return False
# ============================================================
#    def is_only_cancel_mode(self):
#        '''Возможна только кнопка отмены последнего действия'''
#        _fsm_state = self.FSM.GetStateID()
#        if self.is_upgrade_units_waiting: # потому что ожидание юнита для апгрейда в события запихнуть проблематично
#            return True #cancel_upgrade
#        elif _fsm_state in ['repair_wait', 'build_wait']:
#            return True   #cancel_repair
#        return False
# ============================================================
    def Get_menu_items_id(self):
        '''Возвращает список ID кнопок для PopupMenu'''
        if self.cached_menu_items == None:
            self.cached_menu_items = self._Get_menu_items_id_int()
        return self.cached_menu_items
# ============================================================
    def _Get_menu_items_id_int(self):
        '''Возвращает список ID кнопок для PopupMenu'''
        result = []
        _is_clickable, _extra = self.is_clickable(True)
        if not _is_clickable:
            if _extra == 'barbarian_attack':
                _is_no_button_state = self.is_no_button_state()
                if _is_no_button_state: return result
            else:
                return result
        #if not _is_clickable: return result    # early check
        _fsm_state = self.FSM.GetStateID()
        if self.is_wait_worker_state(_fsm_state): return result # early check
        _need_repair, _extra = self._need_repair_button(True, True)
        if _need_repair:
            return ['repair']
        _need_upgrade, _extra = self._need_upgrade_button(True, True)
        if _need_upgrade:
            result.append('upgrade')
        # check for workers >>>
        #_is_not_enough_workers_for_actions = self.is_not_enough_workers_for_actions()
        #if _is_not_enough_workers_for_actions:
        #    del result[:]
        # <<<<<
        _need_restore, _extra = self._need_restore_eff_button(True, True)
        if _need_restore:
            result.append('restore_eff')
        _need_remove, _extra = self._need_remove_button(True)
        if _need_remove:
            result.append('remove')
        return result
# ============================================================
    _menu_id_to_textid = {'upgrade': 'TIPS_UPGRADE',
                 'restore_eff': 'TIPS_RESTORING',
                 'remove': 'TIPS_DEMOLITION',
                 'repair': 'TIPS_REPAIR'}
    def Get_menu_item_text(self, id, over, measure = False):
        _text = self._menu_id_to_textid[id]
        _is_red = False
        _result = []
        _result.append( ('text', ['COLOR_GREEN', 'COLOR_YELLOW3'][over == True]) )
        if id in ['upgrade', 'repair']:
            if measure:
                mresult = [('text', 'COLOR_GREEN')]
                _res = self._get_willgive_res()
                _extra_result = self._need_upgrade_button_extra(['', 'after_next_level_name'][id == 'upgrade'])
                if _extra_result[0] == True:
                    #mresult.append( ('text', 'TIPS_UPGRADE_IMPOSSIBLE') )
                    #mresult.append( ('text', 'TIPS_MENU_NOT_ENOUGH_WORKERS') )
                    if id == 'upgrade':
                        mresult.append( ('text', 'TIPS_MENU_UPGRADE_MEASURE') )
                    elif id == 'repair':
                        mresult.append( ('text', 'TIPS_MENU_REPAIR_MEASURE') )
                elif _res != None and id == 'upgrade':
                    mresult.append( ('text', 'TIPS_UPGRADE') )
                    self._append_menu_willgive(mresult)
                else:
                    #mresult.append( ('text', 'TIPS_UPGRADE_IMPOSSIBLE') )
                    if id == 'upgrade':
                        mresult.append( ('text', 'TIPS_MENU_UPGRADE_MEASURE') )
                    elif id == 'repair':
                        mresult.append( ('text', 'TIPS_MENU_REPAIR_MEASURE') )
                return mresult
            if id == 'upgrade':
                _need_upgrade, _extra = self._need_upgrade_button(True)
                if _need_upgrade:
                    _is_gray = _extra.startswith('gray')
                    if _is_gray:
                        _result = [('text', 'COLOR_GREEN') ]
                        _result.append( ('text', 'TIPS_UPGRADE_IMPOSSIBLE') )
                        return _result
                    _is_red = _extra.startswith('red')
                    if _is_red:
                        _result = [('text', 'COLOR_GREEN')]
                        _result.append( ('text', 'TIPS_UPGRADE_IMPOSSIBLE') )
                        return _result
                if not self.button_res_enough:
                    _is_red = True
                    _result = [('text', 'COLOR_GREEN')]
                    _result.append( ('text', 'TIPS_UPGRADE') )
                    self._append_menu_willgive(_result, over, _is_red)
                    return _result
            elif id == 'repair':
                _need_repair, _extra = self._need_repair_button(True)
                if _need_repair:
                    _is_gray = _extra.startswith('gray')
                    if _is_gray:
                        _result = [('text', 'COLOR_GREEN')]
                        _result.append( ('text', 'TIPS_REPAIR_IMPOSSIBLE') )
                        return _result
                    _is_red = _extra.startswith('red')
                    if _is_red:
                        _result = [('text', 'COLOR_GREEN')]
                        _result.append( ('text', 'TIPS_REPAIR_IMPOSSIBLE') )
                        return _result
                if not self.button_res_enough:
                    _result = [('text', 'COLOR_GREEN')]
                    _result.append( ('text', 'TIPS_REPAIR') )
                    return _result
            if over:
                if not self.button_res_enough:
                    diff_res = UStats.GetNotEnoughRes(self.next_res)
                    if diff_res:
                        _result.append( ('text', 'COLOR_GREEN') )
        elif id == 'restore_eff':
            if measure:
                return [('text', 'COLOR_GREEN'), ('text', 'TIPS_MENU_RESTORE_MEASURE')]
            _restore_res = self.get_res('restore_resname', 'restore_res')
            _need_restore, _extra = self._need_restore_eff_button(True)
            if _need_restore:
                if _extra == 'gray':
                    _result = [('text', 'COLOR_GREEN')]    #COLOR_RED
                    _result.append( ('text', 'TIPS_RESTORING_IMPOSSIBLE') )
                    return _result
            if over:
                if _restore_res:
                    _is_res_enough = UStats.IsResEnough(_restore_res)
                    if not _is_res_enough:
                        diff_res = UStats.GetNotEnoughRes(_restore_res)
                        if diff_res:
                            _result.append( ('text', 'COLOR_GREEN') )
        elif id == 'remove':
            if measure:
                return [('text', 'COLOR_GREEN'), ('text', 'TIPS_MENU_REMOVE_MEASURE')]
            _need_remove, _extra = self._need_remove_button(True)
            _is_gray = _extra.startswith('gray')
            if not _need_remove or _is_gray:
                _result = [('text', 'COLOR_GREEN')]  #COLOR_RED
                _result.append( ('text', 'TIPS_DEMOLITION_IMPOSSIBLE') )
                return _result
        _result.append( ('text', _text) )
        if id == 'upgrade':
            self._append_menu_willgive(_result, over, _is_red)
        return _result
# ============================================================
    def _append_menu_willgive(self, result, over = False, red = False):
        _res = self._get_willgive_res()
        if _res:
            _r = []
            _r.append('space')
            _r.append( ('text', 'PARENTHESIS_LEFT') )
            _r.append( ('text', 'WILL_GIVE') )
            _r.append('space')
            _res_numbers = self._append_res_values(_res)
            if not _res_numbers:
                return
            _r.extend( _res_numbers )
            _r.append( ('text', 'PARENTHESIS_RIGHT') )
            result.extend(_r)
# ============================================================
    def Get_menu_item_res_text(self, id, over):
        result = [('text', ('COLOR_BROWN2', 'COLOR_WHITE')[over == True])]
        _over = over
        _is_clickable, _extra = self.is_clickable()
        if not _is_clickable and _extra == 'barbarian_attack':
            result = [('text', 'COLOR_RED')]# при атаке варварами покажем спец. текст для апгрейда
            result.append( ('text', 'ON_BARB_ATTACK_IMPOSSIBLE') )
            return result
        if id in ['upgrade', 'repair']:
            if id == 'upgrade':
                _need_upgrade, _extra = self._need_upgrade_button(True)
                if _need_upgrade:
                    _is_gray = _extra.startswith('gray')
                    if _is_gray: result = [ ('text', 'COLOR_BROWN2') ]
                    if _extra == 'gray':
                        result.append( ('text', 'TIPS_MENU_CANNOT_DO_IT_YET') )
                        return result
                    elif _extra == 'gray_fields':
                        result.append( ('text', 'TIPS_MENU_NO_FIELD_PLACE') )
                        return result
                    elif _extra == 'red_workers':
                        result = [('text', 'COLOR_RED')]
                        result.append( ('text', 'TIPS_MENU_NOT_ENOUGH_WORKERS') )
                        return result
            elif id == 'repair':
                _need_repair, _extra = self._need_repair_button(True)
                if _need_repair:
                    _is_gray = _extra.startswith('gray')
                    if _is_gray: result = [ ('text', 'COLOR_BROWN2') ]
                    if _extra == 'gray':
                        result.append( ('text', 'TIPS_MENU_CANNOT_DO_IT_YET') )
                        return result
                    elif _extra == 'red_workers':
                        result = [('text', 'COLOR_RED')]
                        result.append( ('text', 'TIPS_MENU_NOT_ENOUGH_WORKERS') )
                        return result
            if _over:
                if not self.button_res_enough:
                    diff_res = UStats.GetNotEnoughRes(self.next_res)
                    if diff_res:
                        result = [('text', 'COLOR_RED')]
                        result.append( ('text', 'NOT_ENOUGH') )
                        result.append( 'space' )
                        res_numbers = self._append_res_values(diff_res, False)
                        #result.append( ('text', 'COLOR_RED') )
                        result.extend( res_numbers )
                else:
                    _over = False
            if not _over:
                result.append( ('text', ['COLOR_BROWN2','COLOR_WHITE'][over == True]) )
                result.append( ('text', 'NEED_FOR_ACTION') )
                _res = self.next_res
                if _res:
                    res_numbers = self._append_res_values(_res, False)
                    result.append('space')
                    result.append( ('text', ['COLOR_BROWN2','COLOR_WHITE'][over == True]) )
                    result.extend(res_numbers)
        elif id == 'remove':
            _return_res = self.get_res('remove_resname', 'remove_res')
            _need_remove, _extra = self._need_remove_button(True)
            _is_gray = _extra.startswith('gray')
            if _is_gray: result.append( ('text', 'COLOR_BROWN2') )
            if _need_remove and _extra == 'gray_personal':
                _has_remove_text = self.HasAttr('remove_cannot_text')
                if _has_remove_text:
                    result.append( ('text', self.GetAttrStr('remove_cannot_text')) )
                else:
                    result.append( ('text', 'REMOVE_CANNOT') )
                return result
            if _is_gray:
                result.append( ('text', 'TIPS_MENU_CANNOT_DO_IT_YET') )
            else:
                result.append( ('text', 'RETURN') )
                if _return_res:
                    res_numbers = self._append_res_values(_return_res, False)
                    result.append('space')
                    result.append( ('text', ['COLOR_BROWN2','COLOR_WHITE'][over == True]) )
                    result.extend(res_numbers)
        elif id == 'restore_eff':
            _restore_res = self.get_res('restore_resname', 'restore_res')
            _need_restore, _extra = self._need_restore_eff_button(True)
            if _need_restore:
                _is_gray = _extra.startswith('gray')
                if _is_gray:
                    result.append( ('text', 'COLOR_BROWN2') )
                    if _extra == 'gray':
                        result.append( ('text', 'TIPS_MENU_FOR_EFFECTIVENESS_50') )
                        return result
                    elif _extra == 'gray_blocked':
                        result.append( ('text', 'TIPS_MENU_CANNOT_DO_IT_YET') )
                        return result
                    else:
                        trace('-- ERROR RESTORE GRAY --')
                        result = []
                        return result
            if _over:
                if _restore_res:
                    _is_res_enough = UStats.IsResEnough(_restore_res)
                    if not _is_res_enough:
                        diff_res = UStats.GetNotEnoughRes(_restore_res)
                        if diff_res:
                            result.append( ('text', 'COLOR_RED') )
                            result.append( ('text', 'NOT_ENOUGH') )
                            result.append( 'space' )
                            res_numbers = self._append_res_values(diff_res, False)
                            result.extend( res_numbers )
                    else:
                        _over = False
            if not _over:
                if _restore_res:
                    result.append( ('text', ['COLOR_BROWN2','COLOR_WHITE'][over == True]) )
                    result.append( ('text', 'NEED_FOR_ACTION') )
                    res_numbers = self._append_res_values(_restore_res, False)
                    result.append('space')
                    result.append( ('text', ['COLOR_BROWN2','COLOR_WHITE'][over == True]) )
                    result.extend(res_numbers)
        return result
# ============================================================
    remove_restore_states = {'prenormal':0, 'normal':0, 'thrown':0,
                        'generating':0, 'generated':0}
    def has_remove_or_restore(self, _fsm_state):
        #_fsm_state = self.FSM.GetStateID()
        if _fsm_state in self.remove_restore_states: #, 'wait_res_unit':0}:
            _need_remove, _extra = self._need_remove_button(True)
            if _need_remove:
                return True
            _need_restore, _extra = self._need_restore_eff_button(True, True)
            if _need_restore:
                return True
        if _fsm_state == 'preremoving': # можно кликнуть для отмены
            return True
        return False
# ============================================================
    def Is_menu_item_green(self, id):
        _is_clickable, _extra = self.is_clickable()
        if not _is_clickable and _extra == 'barbarian_attack':
            return 0
        if id == 'repair':
            _need_repair, _extra = self._need_repair_button(False)
            if not _need_repair:
                return 0
            elif _extra.startswith('gray'):
                return 2
            elif _extra.startswith('red'):
                return 0
        elif id == 'upgrade':
            _need_upgrade, _extra = self._need_upgrade_button(False)
            if not _need_upgrade:
                #if _extra == 'barbarian_attack': return 1
                #else: return 0
                return 0
            elif _extra.startswith('gray'):
                return 2
            elif _extra.startswith('red'):
                return 0
        elif id == 'restore_eff':
            _need_restore, _extra = self._need_restore_eff_button()
            if not _need_restore:
                return 0
            elif _extra.startswith('gray'):
                return 2
        elif id == 'remove':
            _need_remove, _extra = self._need_remove_button()
            if not _need_remove:
                return 0
            elif _extra.startswith('gray'):
                return 2
        return 1
# ============================================================
    def _need_remove_button(self, prechecked = False):
        if self.is_indians:
            return (False, '')
        if not prechecked:
            _is_clickable, _extra = self.is_clickable()
            if not _is_clickable:
                return (False, _extra)
            _fsm_state = self.FSM.GetStateID()
            if self.is_wait_worker_state(_fsm_state): return (False, '')
        if self.cached_need_remove_btn == None:
            self.cached_need_remove_btn = self._need_remove_button_int()
        return self.cached_need_remove_btn
# ============================================================
    def _need_remove_button_int(self):
        _is_broken = self.FSM.GetStateID() == 'broken'
        if not _is_broken:
            _BLOCK_REMOVE = UStats.GetLevelValue('BLOCK_REMOVE')
            if _BLOCK_REMOVE > 0 and not self.unblock_remove:
                return (False, '')
            _can_remove = self.GetAttrBool('can_remove')
            if _can_remove:
                if self.blocked_remove:
                    return (True, 'gray')
                _has_priorities = BMgr.HasPriorities(self.get('locind'))
                if _has_priorities and not self.unblock_remove: # пока есть принуждения - запрещаем уничтожать (чтобы квесты выполнились)
                    return (True, 'gray')
                return (True, '')
            else:
                #if self.is_indians:
                #    return (False, '')
                _has_remove_text = self.HasAttr('remove_cannot_text')
                if _has_remove_text:
                    return (True, 'gray_personal')
        return (False, '')
# ============================================================
    def _need_upgrade_button(self, prechecked = False, skipResEnough = False):
        if self.is_indians:
            return (False, '')
        if not prechecked:
            _is_clickable, _extra = self.is_clickable()
            if not _is_clickable:
                return (False, _extra)
            _fsm_state = self.FSM.GetStateID()
            if self.is_wait_worker_state(_fsm_state):
                return (False, '')
        if skipResEnough:
            if self.cached_need_update_skip_btn == None:
                self.cached_need_update_skip_btn = self._need_upgrade_button_int(skipResEnough)
            return self.cached_need_update_skip_btn
        else:
            if self.cached_need_update_noskip_btn == None:
                self.cached_need_update_noskip_btn = self._need_upgrade_button_int(skipResEnough)
            return self.cached_need_update_noskip_btn
# ============================================================
    def _need_upgrade_button_int(self, skipResEnough = False):
        _fsm_state = self.FSM.GetStateID()
        if _fsm_state != 'broken':
            state = self.SM.GetCurrentState()
            if state not in self._upgrade__build and self.status != 'place':
                _next_name = self.get_next_level_name(True)
                if not _next_name: return (False, '')
                result = self._need_upgrade_button_extra('after_next_level_name')
                if result[0] == True: return result[1]
                _is_priority_blocked = self.is_priority_blocked()
                if _is_priority_blocked:
                    return (True, 'gray')
                _is_next_level_possible = self.is_next_level_possible(skipResEnough)
                if _is_next_level_possible:
                    _name = self.get('name')
                    _need_worker = not self.GetAttrBool('need_no_worker')
                    map_index = self.get('locind')
                    _not_enough_workers = not BMgr.Is_build_action_possible(map_index, _name, _need_worker)
                    if _not_enough_workers:
                        return (True, 'red_workers')
                    return (True, '')
        return (False, '')
# ============================================================
    def _need_upgrade_button_extra(self, place):
        return [False]
# ============================================================
    wc_quarry_stypes = ['woodcutter', 'quarry']
    def _need_restore_eff_button(self, prechecked = False, skipResEnough = False):
        if self.is_indians:
            return (False, '')
        _stype = self.get('stype')
        if _stype in self.has_no_effectiveness:
            return (False, '')
        if _stype in self.wc_quarry_stypes:
            return (False, '')
        _BLOCK_RESTORE_EFF = UStats.GetLevelValue('BLOCK_RESTORE_EFF')
        if _BLOCK_RESTORE_EFF > 0 and not self.unblock_restore_eff:
            return (False, '')
# -----------
        if not prechecked:
            _is_clickable, _extra = self.is_clickable()
            if not _is_clickable: return (False, _extra)
            _fsm_state = self.FSM.GetStateID()
            if self.is_wait_worker_state(_fsm_state): return (False, '')
# -----------
        _cur_eff = self.get_current_lowest_effectiveness()
        if _cur_eff > 50: # 50 # ранее не позволяем восстанавливать эффективность
            if _cur_eff < 100:
                return (True, 'gray')
            return (False, '')
        elif not skipResEnough:
            _is_res_enough = UStats.IsResEnough(self.GetRes('restore_res'))
            if not _is_res_enough:
                return (False, '')
        return (True, '')
# ============================================================
    def _need_repair_button(self, prechecked = False, skipResEnough = False):
        _fsm_state = self.FSM.GetStateID()
        if not prechecked:
            _is_clickable, _extra = self.is_clickable(True)
            if not _is_clickable:
                return (False, _extra)
            if self.is_wait_worker_state(_fsm_state):
                return (False, '')
        _is_next_level_possible = self.is_next_level_possible(skipResEnough)
        state = self.SM.GetCurrentState()
        if _fsm_state == 'broken':
            if state not in self._upgrade__build and self.status != 'place':
                _next_name = self.get_next_level_name(True)
                if not _next_name: return (False, '')
                _is_priority_blocked = self.is_priority_blocked()
                if _is_priority_blocked:
                    return (True, 'gray')
                if _is_next_level_possible:
                    _name = self.get('name')
                    _need_worker = not self.GetAttrBool('need_no_worker')
                    map_index = self.get('locind')
                    _not_enough_workers = not BMgr.Is_build_action_possible(map_index, _name, _need_worker)
                    if _not_enough_workers:
                        return (True, 'red_workers')
                    return (True, '')
        return (False, '')
# ============================================================
    def On_menu_item_click(self, id):
        _is_green = self.Is_menu_item_green(id)
        if _is_green == 1:
            _r = False
            if id == 'remove':
                self.FSM.HandleEvent('remove_event')
                _r = True
            elif id == 'upgrade':
                _r = self.try_start_upgrade_repair('default')
            elif id == 'repair':
                _r = self.try_start_upgrade_repair('default')
            elif id == 'restore_eff':
                _r = self.restore_eff()
            if _r: self.PlaySoundG('SND_CLICK_TASK')    #'SND_BTN_CLICK_PAPER')
        else:
            self.PlaySoundG('SND_ACTION_WRONG')
        if _is_green == 1:
            UI.HidePopup()
# ============================================================
    def restore_eff(self):
        _need_restore_res = self.GetRes('restore_res')
        _is_res_enough = UStats.IsResEnough(_need_restore_res)
        if not _is_res_enough:
            return False
        UStats.SubRes(_need_restore_res)
        if self.get_effective_gen_time(100) != 0:
            self.restore_eff_k = self.get_effective_gen_time() / self.get_effective_gen_time(100)
        else:
            self.restore_eff_k = 1
        self.current_effectiveness = 100
        trace('CALL: RestoreEff', self.get('uid'), self.get('name'))
        EMgr.SendGlobalMsg('effectiveness_changed', '')
        self.reinit_effectiveness_params(True)
        EMgr.send_event('building_restore_eff', self)
        self.PlaySound('SND_EFFICIENCY_UP')
        self.show_arrow(False)
        self.Drawer.ShowImages('restore_eff')
        self.Drawer.StartEffectOnce('restore_eff')
        if True:
            letter = []
            letter.append( ('text', 'MSG_RESTORED_EFF') )
            letter.append( 'space' )
            _loc_ind = self.get('locind')
            EMgr.AppendLetter(letter, self.get('uid'), _loc_ind)
        if LBoard.Core.episode == 1 and LBoard.Core.level == 3:
            _need_help = LBoard.Core.is_help_needed(9)
            if _need_help:
                _uid = self.get('uid')
                if _uid == '3826':
                    Camera.MoveTo(self)
                    UI.open_help(9)
        TMgr.tutHelper.refresh()
        return True
# ============================================================
    def get_popup_position(self, cur_pos, tip_size):
        return toolz.Tools.get_popup_position(cur_pos, tip_size)
# ============================================================
    def do_game_action(self, data):
        '''type, txt_param1, txt_param2, int_param1, int_param2, is_offscreen'''
        TMgr.do_game_action(self, data)
# ============================================================
    def do_action(self, name):
        _action_done = self._do_action(self, name)  # for overriding
        if not _action_done:
            if name == '':
                pass
# ============================================================
    def _do_action(self, name):
        '''For overriding in child classes'''
        return False
# ============================================================
    def get_prev_levelnames(self):
        '''Применимо только для названия варинта level_'''
        _variants = {'level_1':[], 'level_2':['level_1'], 'level_3':['level_1','level_2']}
        _cur_vname = self.get('variant_name')
        if _cur_vname in _variants:
            return _variants[_cur_vname]
        else:
            return []
# ============================================================
    def get_next_levelnames(self):
        '''Применимо только для названия варинта level_'''
        _variants = {'level_1':['level_2', 'level_3'], 'level_2':['level_3'], 'level_3':[]}
        _cur_vname = self.get('variant_name')
        if _cur_vname in _variants:
            return _variants[_cur_vname]
        else:
            return []
# ============================================================
# ============================================================
    def init_fsm(self):
        FSM = self.FSM

        self.UseFSM(True)
        FSM.Init('init')
        _states = [fsm_state.StateInit, StatePrenormal, fsm_state.StateNormal, StateWaitUnit,
                    StateAttacked, StateBroken, #StateWaitResUnit,
                    StateGenerating, StateGenerated,
                    StateUpgrading, StateRepairWait, StateRepairing,
                    fsm_state.StateBuildWait, fsm_state.StateBuilding, fsm_state.StateSurvived,
                    fsm_state.StateThrown,fsm_state.StateMapPlace,
                    fsm_state.StatePreremoving, fsm_state.StateRemoving, fsm_state.StateRemovingFinal]
        for state in _states: self.states[state.name] = state(self)

        #FSM.AddTransition('load_init_event', 'init', 'init')
        #FSM.AddTransition('wait_unit_event', 'normal', 'wait_unit')
        FSM.AddTransition('wait_unit_event', 'prenormal', 'wait_unit')  #\/

        FSM.AddTransition('unit_in_event', 'wait_unit', 'prenormal')    #\/
        #FSM.AddTransition('unit_in_event', 'attacked', 'attacked')
        #FSM.AddTransition('unit_in_event', 'broken', 'broken')

        #FSM.AddTransition('wait_res_unit_event', 'prenormal', 'wait_res_unit')  #\/
        #FSM.AddTransition('res_unit_in_event', 'wait_res_unit', 'generating')   #\/
        FSM.AddTransition('res_unit_in_event', 'prenormal', 'generating')   #\/
        FSM.AddTransition('res_unit_in_event', 'normal', 'normal')   #\/
        FSM.AddTransition('res_unit_in_event', 'wait_unit', 'wait_unit')   #\/

        #FSM.AddTransition('start_generation_event', 'normal', 'generating')
        FSM.AddTransition('start_generation_event', 'prenormal', 'generating')  #\/

        FSM.AddTransition('generated_event', 'generating', 'generated')         #\/
        FSM.AddTransition('res_taken_event', 'generated', 'prenormal')          #\/

        FSM.AddTransition('res_taken_event', 'prenormal', 'prenormal')          #\/
        FSM.AddTransition('try_again_event', 'prenormal', 'prenormal')          #\/

        FSM.AddTransition('attack_event', 'generating', 'attacked')     #\/
        FSM.AddTransition('attack_event', 'normal', 'attacked')         #\/
        FSM.AddTransition('attack_event', 'generated', 'attacked')      #\/
        FSM.AddTransition('attack_event', 'prenormal', 'attacked')      #\/
        FSM.AddTransition('attack_event', 'wait_unit', 'attacked')      #\/
        #FSM.AddTransition('attack_event', 'wait_res_unit', 'attacked')  #\/

        FSM.AddTransition('attack_stop_event', 'attacked', 'prenormal')
        FSM.AddTransition('repair_wait_event', 'broken', 'repair_wait') #\/
        FSM.AddTransition('repair_event', 'repair_wait', 'repairing')   #\/

        FSM.AddTransition('upgrade_event', 'normal', 'upgrading')   #\/
        FSM.AddTransition('upgrade_event', 'prenormal', 'upgrading')#\/
        FSM.AddTransition('upgrade_event', 'generating', 'upgrading')#\/
        FSM.AddTransition('upgrade_event', 'generated', 'upgrading')#\/
        FSM.AddTransition('upgrade_event', 'generated', 'upgrading')#\/
        #FSM.AddTransition('upgrade_event', 'wait_res_unit', 'upgrading')#\/
        FSM.AddTransition('upgrade_event', 'wait_unit', 'upgrading')#\/

        FSM.AddTransition('remove_event', 'prenormal', 'preremoving')
        FSM.AddTransition('remove_event', 'normal', 'preremoving')
        FSM.AddTransition('remove_event', 'generating', 'preremoving')
        FSM.AddTransition('remove_event', 'generated', 'preremoving')
        #FSM.AddTransition('remove_event', 'wait_res_unit', 'preremoving')
        FSM.AddTransition('remove_event', 'thrown', 'preremoving')
        FSM.AddTransition('removing_event', 'preremoving', 'removing')
        FSM.AddTransition('removing_final_event', 'removing', 'removing_final')
        FSM.AddTransition('removed_event', 'removing_final', '_dead_')

        FSM.AddTransition('build_wait_event', 'init', 'build_wait')     #\/
        FSM.AddTransition('build_wait_event', 'map_place', 'build_wait')#\/
        FSM.AddTransition('build_event', 'build_wait', 'building')      #\/

        FSM.AddTransition('broken_event', 'init', 'broken')         #\/
        FSM.AddTransition('prenormal_event', 'init', 'prenormal')   #\/
        FSM.AddTransition('map_place_event', 'init', 'map_place')   #\/
        FSM.AddTransition('survived_event', 'init', 'survived')     #\/
        FSM.AddTransition('thrown_event', 'init', 'thrown')     #\/

        FSM.AddTransition('prenormal_event', 'repairing', 'prenormal')  #\/
        FSM.AddTransition('prenormal_event', 'upgrading', 'prenormal')  #\/
        FSM.AddTransition('prenormal_event', 'building', 'prenormal')   #\/

        #FSM.AddTransition('break_event', 'normal', 'broken')
        #FSM.AddTransition('break_event', 'generating', 'broken')
        #FSM.AddTransition('break_event', 'generated', 'broken')
        FSM.AddTransition('break_event', 'attacked', 'broken')

        FSM.AddTransition('normal_event', 'repairing', 'normal')    #\/
        FSM.AddTransition('normal_event', 'upgrading', 'normal')    #\/
        FSM.AddTransition('normal_event', 'building', 'normal')     #\/
        FSM.AddTransition('normal_event', 'prenormal', 'normal')    #\/

        FSM.AddTransition('thrown_event', 'normal', 'thrown')       #\/
        FSM.AddTransition('thrown_event', 'prenormal', 'thrown')    #\/
        FSM.AddTransition('thrown_event', 'generating', 'thrown')   #\/
        FSM.AddTransition('thrown_event', 'generated', 'thrown')    #\/
        FSM.AddTransition('thrown_event', 'preremoving', 'thrown')    #\/
        FSM.AddTransition('thrown_event', 'wait_unit', 'thrown')    #\/

        FSM.AddTransition('survived_saved_event', 'survived', 'prenormal')  #\/
        FSM.AddTransition('cancel_mapplace_event', 'build_wait', 'map_place')#\/
        FSM.AddTransition('cancel_event', 'build_wait', '_dead_')           #\/
        FSM.AddTransition('cancel_event', 'repair_wait', 'broken')          #\/
        FSM.AddTransition('cancel_event', 'preremoving', 'prenormal')       #\/
# ============================================================
    def fsm_on_can_enter(self, next_state, event_id):   # :bool
        if next_state in self.states:
            state = self.states[next_state]
            return state.can_enter(event_id)
        return True
# ============================================================
    def fsm_on_can_exit(self, cur_state, event_id):     # :bool
        if cur_state in self.states:
            state = self.states[cur_state]
            return state.can_exit(event_id)
        return True
# ============================================================
    def fsm_on_enter(self, next_state, event_id):
        if next_state in self.states:
            state = self.states[next_state]
            state.on_enter(event_id)
# ============================================================
    def fsm_on_exit(self, cur_state, next_state):
        if cur_state in self.states:
            state = self.states[cur_state]
            state.on_exit(next_state)
# ============================================================
    def fsm_on_handle(self, cur_state, event_id):
        if cur_state in self.states:
            state = self.states[cur_state]
            state.on_handle(event_id)
# ============================================================
    def fsm_on_state_changed(self):
        cur_state = self.FSM.GetStateID()
        if cur_state in self.states:
            state = self.states[cur_state]
            state.on_state_changed()
        self.switch_animation()
# ============================================================
# ============================================================
    def DEBUG_get_info(self):
        r = []
        self.DEBUG_b_info(r)
        _state_name = self.FSM.GetStateID()
        r.append(('debug', 'S: '))
        r.append(('debug', _state_name))
        r.append('newline')
        r.append( ('debug', str(self.get_effective_gen_time()) + ';' + str(self.current_effectiveness) + ';' + str(self.restore_eff_k) ))
        if self.tasks_member:
            for tname in self.tasks_member:
                r.append( 'newline' )
                r.append( ('debug', tname) )
                t = self.tasks_member[tname][1]
                r.append( 'space' )
                r.append( ('digit', t.uid) )
        return r
# ============================================================
    def DEBUG_b_info(self, r):
        r.append(('debug', self.get('uid')))
        r.append('space')
        r.append( ('debug', 'bres:') )
        r.append( ('digit', self.has_base_res) )
        r.append('space')
        r.append( ('debug', 'w:') )
        r.append( ('digit', len(self.worker_units)) )
        r.append('space')
        r.append( ('debug', 'in:') )
        r.append( ('digit', len(self._internal_free_units)) )
        r.append('space')
        r.append( ('debug', 'wait:') )
        r.append( ('digit', self.is_wait_res_unit) )
        r.append( 'newline' )
# ============================================================
# ============================================================
# ============================================================
# ============================================================
class StatePrenormal(fsm_state.FSMState):
    name = 'prenormal'

    def __init__(self, item = None):
        fsm_state.FSMState.__init__(self, item)
        self.ser_name = 'StatePrenormal'

    def on_enter(self, evID):
        b = self.item
        if b.is_under_attack:
            trace('>> ScriptFSM: precomplete is_under_attack')
            return    # а что делать-то? Если, конечно мы здесь окажемся

        if b.status == 'broken':
            _vname = b.get('variant_name')
            _is_thrown = _vname.startswith('thrown')
            if _is_thrown:
                b.FSM.HandleEvent('thrown_event')
                return
        _last_fsm_event = ''
        _need_worker = False
        if b.need_call_units():
            b.FSM.HandleEvent('wait_unit_event')             # нужно позвать работника в здание
            _last_fsm_event = 'wait_unit_event'
            _need_worker = True     # чтобы "заказать" ресурс до того, как заселится работник
        #else:   # если уже есть работники
        if True:
            _need_base_resource = b.need_base_resource()
            if _need_base_resource:
                _can_generate = b._can_generate_resource()
                if _can_generate:
                    _is_base_res_enough = b._is_base_res_enough()
                    if _is_base_res_enough:
                        b.FSM.HandleEvent('start_generation_event')  # можно начинать генерировать
                        _last_fsm_event = 'start_generation_event'
                    else:
                        #b.FSM.HandleEvent('wait_res_unit_event') # нужно запросить ресурсы
                        if not b.is_wait_res_unit:  # если еще не ждем ресурсов
                            if _need_worker:
                                b.Signals.Restart('call_unit_with_res', 500)
                                _last_fsm_event = 'call_unit_with_res'
                            else:
                                b._call_unit_with_res()
                                _last_fsm_event = 'F call_unit_with_res'
                        else:
                            _last_fsm_event = 'waiting_res_unit'
                else:
                    if not _need_worker:
                        if b.is_generated_res:
                            b.call_take_res_unit()
                            _last_fsm_event = 'F call_take_res_unit'
                        else:
                            b.FSM.HandleEvent('normal_event')    # а что еще можно сделать в этом случае? Пусть лучше уж просто стоит.
                            _last_fsm_event = 'normal_event'
                    else:
                        _last_fsm_event = 'need_worker'
            else:
                if not _need_worker:
                    _can_generate = b._can_generate_resource()
                    if _can_generate:
                        b.FSM.HandleEvent('start_generation_event')  # можно начинать генерировать
                        _last_fsm_event = 'start_generation_event 2'
                    elif b.is_generated_res:
                        b.call_take_res_unit()
                        _last_fsm_event = 'F call_take_res_unit'
                    else:
                        b.FSM.HandleEvent('normal_event')    # а что еще можно сделать в этом случае? Пусть лучше уж просто стоит.
                        _last_fsm_event = 'normal_event'
                else:
                    _last_fsm_event = 'need_worker 2'
        trace('PRENORMAL result =', _last_fsm_event)

    def can_exit(self, evID):
        if evID == 'res_unit_in_event':
            _units_in = len(self.item.worker_units)
            if _units_in > 0:
                return True
            return False
        #elif evID == 'unit_in_event':
        #    return True
        return True

    def on_handle(self, evID):
        if evID == 'res_unit_in_event':
            self.item.is_wait_res_unit = False
            b = self.item
            _unit = b.res_unit
            if _unit:
                b.take_resource(_unit.resource) # передаем ресурс
                _unit.resource = None
                #mbox(b.get('name'))
# ============================================================
# ============================================================
class StateWaitUnit(fsm_state.FSMState):
    name = 'wait_unit'

    def __init__(self, item = None):
        fsm_state.FSMState.__init__(self, item)
        self.ser_name = 'StateWaitUnit'

    def on_enter(self, evID):
        self.item.call_units()

    def on_handle(self, evID):
        if evID == 'res_unit_in_event':
            self.item.is_wait_res_unit = False
            b = self.item
            _unit = b.res_unit
            if _unit:
                b.take_resource(_unit.resource) # передаем ресурс
                _unit.resource = None
# ============================================================
# ============================================================
class StateAttacked(fsm_state.FSMState):
    name = 'attacked'

    def __init__(self, item = None):
        fsm_state.FSMState.__init__(self, item)
        self.ser_name = 'StateAttacked'

    def on_enter(self, evID):
        b = self.item
        b.on_attack_started()       # начали атаку здания
        b.set_under_attack()
        b.lost_resource()

    def on_exit(self, next_state):
        b = self.item
        b.reset_under_attack()
        b.task = None
        if next_state != 'broken':
            b.on_attack_canceled()
# ============================================================
# ============================================================
class StateBroken(fsm_state.FSMState):
    name = 'broken'
    def on_broken(self):
        pass
# ============================================================
# ============================================================
class StateGenerating(fsm_state.FSMState):
    name = 'generating'

    def __init__(self, item = None):
        fsm_state.FSMState.__init__(self, item)
        self.ser_name = 'StateGenerating'

    def on_enter(self, evID):
        self.item._start_resource_generation()

    def on_exit(self, next_state):
        if next_state != 'generated':
            self.item.lost_resource()
        self.item.is_generating_res = False  # при выходе всегда сбрасываем
# ============================================================
# ============================================================
class StateGenerated(fsm_state.FSMState):
    name = 'generated'

    def __init__(self, item = None):
        fsm_state.FSMState.__init__(self, item)
        self.ser_name = 'StateGenerated'

    def on_enter(self, evID):
        self.item.generate_resource()

    #def on_exit(self, next_state):
    #    if next_state != '':
    #        self.item.lost_resource()
# ============================================================
# ============================================================
class StateUpgrading(fsm_state.FSMState):
    name = 'upgrading'

    def __init__(self, item = None):
        fsm_state.FSMState.__init__(self, item)
        self.ser_name = 'StateUpgrading'

    def on_enter(self, evID):
        b = self.item
        b.is_upgrade_units_waiting = False
        b.on_upgrade_started()
        BMgr.SendGlobalMsg( 'upgrade_started' )
        b.show_clock(False)  # песочные часы ожидания
        b.SM.SetState('upgrade')
        if b.status == 'normal':
            if b.hack_for_upgrade:
                b.show_unit = True
        b.PlaySoundG('SND_BUILDING_LOOP')
        b.Drawer.RestartEffect('upgrading')
        #b.lost_resource()     # остановим генерацию ресурса, если здание начало апгрейд

    def on_handle(self, evID):
        if evID == 'prenormal_event':
            b = self.item
            b.SM.SetState('normal')
            b.switch_level('upgraded')
            trace('Building upgrade complete')   # переключаемся на следующий вариант (после того, как состояние здания перевели в normal !)

    def on_exit(self, next_state):
        b = self.item
        building_info = {'type':b.get('type'), 'subtype':b.get('subtype'), 'uid':b.get('uid')}
        EMgr.send_event( 'build_upgrade_complete', b, building_info )
        _stype = b.get('stype')
        if _stype not in {}:#{'bridge':0, 'dam':0, 'totem':0, 'totem_stone':0}:
            EMgr.send_event( 'building_upgraded', b )  # для prizes
        b._on_upgrade_complete(building_info)
        BMgr.SendGlobalMsg('upgrade_completed')
# ============================================================
# ============================================================
class StateRepairWait(fsm_state.FSMState):
    name = 'repair_wait'

    def __init__(self, item = None):
        fsm_state.FSMState.__init__(self, item)
        self.ser_name = 'StateRepairWait'

    def on_enter(self, evID):
        b = self.item
        TMgr.add_task( task_upgrade.UpgradeTask(b) )
        b.show_arrow(False)
        b.SM.SetState('awaiting_action')    # added 14 september 2012...
        BMgr.SuspendPriority(b.get('name'))
        BMgr.SuspendBuildTemplate(b.get('name'), True)
        b.is_priority_building = True
        b.on_awaiting_action()

    def on_handle(self, evID):
        if evID == 'cancel_event':
            b = self.item
            task = b.get_participating_task('upgrade_building')
            if task:    # если отменяем апгрейд
                b.PlaySoundG('SND_CLICK_TASK_CANCEL')
                task.cancel()
                #self.is_upgrade_units_waiting = False
            b.SM.SetState('normal')
            BMgr.UnsuspendPriority(b.get('name'))
            BMgr.SuspendBuildTemplate(b.get('name'), False)
            self.is_priority_building = False
# ============================================================
# ============================================================
class StateRepairing(fsm_state.FSMState):
    name = 'repairing'

    def __init__(self, item = None):
        fsm_state.FSMState.__init__(self, item)
        self.ser_name = 'StateRepairing'

    def on_enter(self, evID):
        b = self.item
        b.on_upgrade_started()  #@ in repair???
        BMgr.SendGlobalMsg( 'repair_started' )
        b.show_clock(False)  # песочные часы ожидания
        b.SM.SetState('upgrade')
        # ^ upgrade (not working :`( ):
        #self.item.PlaySoundG('SND_BUILDING_LOOP')
        #self.item.Drawer.RestartEffect('repair')
        if b.status == 'normal':
            if b.hack_for_upgrade:
                b.show_unit = True
        b.PlaySoundG('SND_BUILDING_LOOP')
        b.Drawer.RestartEffect('repair')

    def on_handle(self, evID):
        if evID == 'prenormal_event':
            b = self.item
            b.SM.SetState('normal')
            b.switch_level('repaired')   # переключаемся на следующий вариант (после того, как состояние здания перевели в normal !)
            trace('Building repair complete')

    def on_exit(self, next_state):
        b = self.item
        building_info = {'type':b.get('type'), 'subtype':b.get('subtype'), 'uid':b.get('uid')}
        EMgr.send_event( 'build_repair_complete', b, building_info )
        _stype = b.get('stype')
        if _stype not in {}:#{'bridge':0, 'dam':0, 'totem':0, 'totem_stone':0}:
            EMgr.send_event( 'building_repaired', b )  # для prizes
        if b.is_priority_building:
            EMgr.send_event('priority_complete', b, building_info)
            b.is_priority_building = False
        b._on_repair_complete(building_info)
        BMgr.SendGlobalMsg('repair_completed')
        BMgr.SendGlobalMsg('buildings_count_change')
# ============================================================
# ============================================================
