# -*- coding: utf-8 -*-

import sys
if not "tinypy" in sys.version:
    from extra import *

import base_unit


class UnitGeneral(base_unit.BaseUnit):
    states_anim_dict = {'hit': 'hit', 'run': 'run', 'stay': 'idle', 'pickup':'pickup',\
                        'magic': 'magic', 'talk':'talk',\
                        'fast_run':'fast_run',
                        'boat_stay':'boat_idle', 'boat_run': 'boat_run' }

    def __init__(self, cdata):
        base_unit.BaseUnit.__init__(self, cdata)
        self.ser_name = 'UnitGeneral'
        self.resource = None
        self.attacked_item = None
        self.attack_state = ''
        self.need_magic = False         # затычка для генерала, чтобы показать магию при взятии gems
        default_unit_speed = Globals.GFloat('DEFAULT_UNIT_SPEED') # default speed
        cur_unit_speed = Globals.GFloat('GENERAL_SPEED')
        self.Drawer.SetAnimBoost('run', 1000*cur_unit_speed/default_unit_speed)
        self.Drawer.SetAnimBoost('fast_run', 1000*cur_unit_speed/default_unit_speed)
        self.Drawer.SetAnimationEvent('boat_run', 30, 'on_boat_move')
# ============================================================
    def on_message(self, msg_type, msg_param1='', msg_param2='', msg_dict={}):
        if msg_type == 'at_target':
            self.SM.SetStateNow('stay')
            # at target
            trace(self.task_type)
            if self.task:
                self.task.target_reached(self)
            return
        elif msg_type == 'hit_item':
            #self.Drawer.PlayAnimationOnce(self.current_action_animation, 'hit_complete')
            self.Drawer.PlayAnimationOnce(self.current_action_animation, 'hit_anim_complete', '', 50, 'hit_complete') # ударим еще раз
        elif msg_type == 'hit_anim_complete':
            trace('general hit_anim_complete')
            self.on_hit_anim_complete()
        elif msg_type == 'hit_complete':
            trace('general hit_complete')
            self.on_hit_complete()
        elif msg_type == 'on_boat_move':
            self.PlaySound('SND_BOAT')
# ============================================================
    def on_signal(self, signal):
        if signal == 'offscreen':
            _state_name = self.SM.GetCurrentState()
            if _state_name in ['stay', '']:   # idle state
                self.set_face_direction()
        elif signal == 'step_aside':
            self.step_aside_now()
        else:
            base_unit.BaseUnit.on_signal(self, signal)
# ============================================================
    def on_click(self, btn_name):
        pass    # на клики не реагирует
# ============================================================
    def on_state_enter(self, state_name, additional):
        self.switch_animation()
# ============================================================
    def on_state_exit(self, state_name, additional):
        if state_name == 'talk':    # завершаем разговор с юнитом
            self.on_talk_complete()
# ============================================================
    def on_state_change(self, cur_state, next_state):
        # у нас, ВРОДЕ, все останавливается при отмене, а здесь только мешает (потому что поздно приходит)
        #if cur_state == "run":
        #    if next_state == "stay":
        #        self.StopMove()
        pass
# ============================================================
    def on_create(self):
        self.SetDirection(self.GetAttrInt('start_direction'))
        self.switch_animation()
# ============================================================
    def __str__(self):
        return base_unit.BaseUnit.__str__(self)
# ============================================================
# ============================================================
# ============================================================
    def switch_animation(self):
        state_name = self.SM.GetCurrentState()
        dir = self.GetDirection()
        self.Drawer.ShowImages('', False)
        self.Drawer.ShowImages('shadow', True)
        new_image_type = 'idle'
        _is_on_boat = self.IsOnBoat()
        if _is_on_boat: # если на лодке, то используем другие анимации:
            if state_name == 'stay': state_name = 'boat_stay'
            if state_name == 'run': state_name = 'boat_run'
            if state_name in ['hit', 'pickup']: state_name = 'boat_stay' # затычка, чтобы сразу анимация менялась
        if state_name == 'run':
            _run_coeff = self.GetRunFinalCoeff()
            if _run_coeff >= 1.25: state_name = 'fast_run'
        if state_name == 'hit':
            new_image_type = self.current_action_animation
        elif state_name in self.states_anim_dict:
            new_image_type = self.states_anim_dict[state_name]

        self.Drawer.ShowImages(new_image_type, True)
        if new_image_type == 'magic':
            #self.Drawer.PlayAnimationOnce('magic', 'magic_complete')        # ударим еще раз
            dir = int(dir / 2)
        self.Drawer.SetSequence(new_image_type, dir)
        #if state_name in ['run', 'fast_run', 'boat_run']:
        _bonus_v = UStats.GetLevelValue('IS_GENERAL_SPEEDBONUS_ACTIVE')
        self.Drawer.ShowImages('bonus', (_bonus_v > 0))
# ============================================================
    # @override
    def stop_walk(self):
        state_name = self.SM.GetCurrentState()
        _need_magic = self.need_magic
        if _need_magic:
            self.show_magic()
        self.StopMove()     # событие сработает не мгновенно, а остановиться нужно сразу!
        if state_name == 'stay':  # уже стоим
            return
        if not _need_magic:
            self.StopMove()
            self.SM.SetState('stay')
            self.switch_animation() # не поломает ли чего? не должно вроде
# ============================================================
    # @override
    def action_hit_item(self, item):
        self.attacked_item = item
        self.SM.SetState('hit')
        self.SendMsg('hit_item')
# ============================================================
    def cancel_hits(self):
        self.attacked_item = None
        self.StopMove()
        self.SM.SetState('stay')
# ============================================================
    def on_hit_complete(self):
        trace('on_hit_complete 1')
        if self.attacked_item:      # проверяем на случай, если отменили задание
            trace('on_hit_complete 2')
            self.attacked_item.play_hit_sound()
            self.attacked_item.action_hit_me()      # удар по объекту
            trace('general hit complete:', self.attacked_item)
            if self.attacked_item:  # в некоторых случаях объект может быть уничтожен
                _is_action_complete = self.attacked_item.is_action_complete()
                if not _is_action_complete:   # объект еще не разрушен
                    trace('on_hit_complete 3')
                    self.attacked_item.scale_bump()     # чтобы объект "дергался"
                else:
                    self.attacked_item.play_complete_sound()
                    trace('on_hit_complete 4')
                    self.attack_state = 'complete'
                    #self.SM.SetState('stay')
                    #self.switch_animation()
                    #self.attacked_item = None
                    #if self.task:
                    #    trace('on_hit_complete 5')
                    #    self.task.subtask.on_action_complete()
# ============================================================
    def on_hit_anim_complete(self):
        if self.attacked_item:      # проверяем на случай, если отменили задание
            if self.attack_state == 'attack' and not self.attacked_item.is_action_complete():   # объект еще не разрушен
                self.Drawer.PlayAnimationOnce(self.current_action_animation, 'hit_anim_complete', '', 50, 'hit_complete') # ударим еще раз
            else:
                self.attack_state = ''
                self.StopMove()
                self.SM.SetState('stay')
                self.switch_animation()
                self.attacked_item = None
                if self.task:
                    self.task.subtask.on_action_complete()
# ============================================================
    def need_show_magic(self):
        self.need_magic = True
# ============================================================
    #@override
    def show_magic(self):
        self.need_magic = False
        self.SM.SetState('magic', 1000)
        self.switch_animation()
# ============================================================
    #@override
    def is_busy(self):
        if self.task != None or self.need_magic or self.unit_is_talking:
            return True
        else:
            return False
# ============================================================
    def step_aside(self):
        self.Signals.Restart('step_aside', 2000)
# ============================================================
    def step_aside_now(self):
        if not self.task:   # если нет задания, то отходим, иначе будут баги
            _cell = self.get('cell')
            trace('step_aside_now center cell:', _cell)
            _destCell = UMgr.GetStepAsideCell( self, _cell, 5 ) # 7 - должно хватить, а то и много
            if _destCell:
                _is_moving = self.IsMoving()
                if not _is_moving:      # проверка, что в данный момент юнит никуда не движется:
                    trace('step_aside_now:', _destCell)
                    self.run_to_target(_destCell)
# ============================================================
    #@override
    def Get_tip_image(self):
        return "IMG_I_TIP_BG"
# ============================================================
    def _get_tips_extra(self, place, result, by_click):
        if place == 'before_access_check_click':
            return False
        return True
# ============================================================
    def Run_to_water_cell(self, row, column):
        '''Отправляет генерала к точке на воде'''
        _cell = {'row':row, 'col':column}
        _cur_cell = self.get('cell')
        if _cell == _cur_cell:  # если клик по той же ячейке
            return
        result = self.TryRunTo(_cell)
        if result == False:
            trace('run_to_target failed', self.get('uid'), _cell)
            return
        #self.show()
        return
# ============================================================
    def on_boat_on(self, _boat):
        self.SetOnBoat(True)
        _boat_cell = _boat.get('cell')
        BMgr.RemoveItemFromMap(_boat)
        self.set('cell', _boat_cell)
        self.ClearGeneralAccess()
        self.CheckGeneralAccess()
# ============================================================
    def On_boat_off(self, _boat_row, _boat_col):
        map_index = self.get('locind')
        self.ClearGeneralAccess()   # сначала чистим доступность - после метода 'On_boat_off' вызывается CheckGeneralAccess()
        _boat_cell = {'row':_boat_row, 'col':_boat_col}
        _boat = BMgr.AddItemToMap(map_index, 'boat1', '', _boat_cell, True)
        if _boat:
            _is_passable = self.IsCellPassable(self.get('cell'))
            if _is_passable:
                _boat.SetAccessible(True)
            else:
                _boat.SetGeneralAccessible(True)
        self.SetOnBoat(False)
        self.CheckGeneralAccess()
# ============================================================
    #@override
    #def do_speaking(self, speak):
    #    self.SM.SetState('talk')
    #    self.switch_animation()
# ============================================================
# ============================================================
# ============================================================


